"""Read MCMCSamples from Cobaya chains."""
import os
import re
import numpy as np
from anesthetic.samples import MCMCSamples
from pandas import concat


def read_paramnames(root):
    """Read header of ``<root>.1.txt`` to infer the paramnames.

    This is the data file of the first chain. It should have as many
    columns as there are parameters (sampled and derived) plus an
    additional two corresponding to the weights (first column) and the
    log-posterior (second column). The first line should start with a # and
    should list the parameter names corresponding to the columns. These
    will be used as handles in the pandas array.
    """
    with open(root + ".1.txt") as f:
        header = f.readline()[1:]
        paramnames = header.split()[2:]
        try:
            from getdist import loadMCSamples
            s = loadMCSamples(file_root=root, no_cache=True)
            labels = {p.name: '$' + p.label + '$' for p in s.paramNames.names}
            for p in paramnames:
                if p == 'minuslogprior':
                    labels.update({p: '$-\\ln\\pi$'})
                elif 'minuslogprior_' in p:
                    sub = p.split('_', maxsplit=1)[-1].lstrip('_')
                    labels.update({p: f'$-\\ln\\pi_\\mathrm{{{sub}}}$'})
            return paramnames, labels
        except ImportError:
            return paramnames, {}


def read_cobaya(root, *args, **kwargs):
    """Read Cobaya yaml files.

    Note that in order to optimally read chains from Cobaya you need to have
    `GetDist <https://getdist.readthedocs.io/en/latest/>`__ installed.

    Parameters
    ----------
    root : str
        root name for reading files in Cobaya format, i.e. the files
        ``<root>.*.txt`` and ``<root>.updated.yaml``.

    Returns
    -------
    :class:`anesthetic.samples.MCMCSamples`

    """
    dirname, basename = os.path.split(root)

    files = os.listdir(os.path.dirname(root))
    regex = re.escape(basename) + r'.([0-9]+)\.txt'
    matches = [re.match(regex, f) for f in files]
    chains_files = [(m.group(1), os.path.join(dirname, m.group(0)))
                    for m in matches if m]
    if not chains_files:
        raise FileNotFoundError(dirname + '/' + regex + " not found.")

    columns, labels = read_paramnames(root)
    columns = kwargs.pop('columns', columns)
    labels = kwargs.pop('labels', labels)
    kwargs['label'] = kwargs.get('label', os.path.basename(root))

    samples = []
    for i, chains_file in chains_files:
        data = np.loadtxt(chains_file)
        weights, minuslogP, data = np.split(data, [1, 2], axis=1)
        mcmc = MCMCSamples(data=data, columns=columns,
                           weights=weights.flatten(),
                           labels=labels, *args, **kwargs)
        mcmc['logP'] = -minuslogP
        mcmc.set_label('logP', '$\\ln\\mathcal{P}$')
        mcmc['logL'] = -mcmc['chi2'] / 2
        mcmc.set_label('logL', '$\\ln\\mathcal{L}$')
        mcmc['chain'] = int(i) if i else np.nan
        samples.append(mcmc)

    samples = concat(samples)
    samples.index.names = ['index', 'weights']
    samples.sort_values(by=['chain', 'index'], inplace=True)
    samples.reset_index(inplace=True, drop=True)
    samples.root = root
    samples.label = kwargs['label']

    if np.all(samples.chain == samples.chain.iloc[0]):
        samples.drop(columns='chain', inplace=True, level=0)
    else:
        samples.set_label('chain', r'$n_\mathrm{chain}$')

    return samples



"""Data-processing utility functions."""
import numpy as np
import pandas
from scipy import special
from scipy.interpolate import interp1d
from scipy.optimize import minimize_scalar
from scipy.stats import kstwobign, entropy
from matplotlib.tri import Triangulation
import contextlib
import inspect
import re


def logsumexp(a, axis=None, b=None, keepdims=False, return_sign=False):
    r"""Compute the log of the sum of exponentials of input elements.

    This function has the same call signature as
    :func:`scipy.special.logsumexp` and mirrors scipy's behaviour except for
    ``-np.inf`` input. If a and b are both ``-inf`` then scipy's function will
    output ``nan`` whereas here we use:

    .. math::

        \lim_{x \to -\infty} x \exp(x) = 0

    Thus, if ``a=-inf`` in ``log(sum(b * exp(a))`` then we can set ``b=0``
    such that that term is ignored in the sum.
    """
    if b is None:
        b = np.ones_like(a)
    b = np.where(a == -np.inf, 0, b)
    return special.logsumexp(a, axis=axis, b=b, keepdims=keepdims,
                             return_sign=return_sign)


def neff(w, beta=1):
    r"""Calculate effective number of samples.

    Using the Huggins-Roy family of effective samples
    (https://aakinshin.net/posts/huggins-roy-ess/).

    Parameters
    ----------
    beta : int, float, str, default = 1
        The value of beta used to calculate the number of effective samples
        according to

        .. math::

            N_{eff} &= \bigg(\sum_{i=0}^n w_i^\beta \bigg)^{\frac{1}{1-\beta}}

            w_i &= \frac{w_i}{\sum_j w_j}

        Beta can take any positive value. Larger beta corresponds to a greater
        compression such that:

        .. math::

            \beta_1 < \beta_2 \Rightarrow N_{eff}(\beta_1) > N_{eff}(\beta_2)

        Alternatively, beta can take one of the following strings as input:

        * If 'inf' or 'equal' is supplied (equivalent to beta=inf), then the
          resulting number of samples is the number of samples when compressed
          to equal weights, and given by:

        .. math::

            w_i &= \frac{w_i}{\sum_j w_j}

            N_{eff} &= \frac{1}{\max_i[w_i]}

        * If 'entropy' is supplied (equivalent to beta=1), then the estimate
          is determined via the entropy based calculation, also referred to as
          the channel capacity:

        .. math::

            H &= -\sum_i p_i \ln p_i

            p_i &= \frac{w_i}{\sum_j w_j}

            N_{eff} &= e^{H}

        * If 'kish' is supplied (equivalent to beta=2), then a Kish estimate
          is computed (Kish, Leslie (1965). Survey Sampling.
          New York: John Wiley & Sons, Inc. ISBN 0-471-10949-5):

        .. math::

            N_{eff} = \frac{(\sum_i w_i)^2}{\sum_i w_i^2}

        * str(float) input gets converted to the corresponding float value.

    """
    w = w / np.sum(w)
    if beta == np.inf or beta == 'inf' or beta == 'equal':
        return 1 / np.max(w)
    elif beta == 'entropy' or beta != 'kish' and str(float(beta)) == '1.0':
        return np.exp(entropy(w))
    else:
        if beta == 'kish':
            beta = 2
        elif isinstance(beta, str):
            beta = float(beta)
        return np.sum(w**beta)**(1/(1-beta))


def compress_weights(w, u=None, ncompress=True):
    """Compresses weights to their approximate channel capacity."""
    if u is None:
        u = np.random.rand(len(w))

    if w is None:
        w = np.ones_like(u)

    if ncompress is True or isinstance(ncompress, str):
        if ncompress is True:
            ncompress = 'entropy'
        ncompress = neff(w, beta=ncompress)
    elif ncompress is False:
        return w

    # TODO: remove this in version >= 2.1
    if ncompress < 0:
        raise ValueError(
            "ncompress<0 is anesthetic 1.0 behaviour. For equally weighted "
            "samples you should now use ncompress='inf' or ncompress='equal'."
            )

    W = w * ncompress / w.sum()

    fraction, integer = np.modf(W)
    integer = integer.astype(int)
    if is_int(ncompress):
        remainder = ncompress - integer.sum()
        mask = fraction > 0
        race_time = np.full_like(fraction, np.inf)  # exp-race arrival time
        race_time[mask] = -np.log(u[mask])/fraction[mask]
        idx = np.argpartition(race_time, remainder-1)[:remainder]
        extra = np.bincount(idx, minlength=len(integer))
    else:
        extra = (u < fraction).astype(int)
    return integer + extra


def quantile(a, q, w=None, interpolation='linear'):
    """Compute the weighted quantile for a one dimensional array."""
    if w is None:
        w = np.ones_like(a)
    a = np.array(list(a))  # Necessary to convert pandas arrays
    w = np.array(list(w))  # Necessary to convert pandas arrays
    i = np.argsort(a)
    c = np.cumsum(w[i[1:]]+w[i[:-1]])
    c = c / c[-1]
    c = np.concatenate(([0.], c))
    icdf = interp1d(c, a[i], kind=interpolation)
    quant = icdf(q)
    if isinstance(q, float):
        quant = float(quant)
    return quant


def var_unbiased(a, w, axis=0, ddof=1):
    """Compute the unbiased variance from weighted samples.

    Uses the standard reliability-weight correction
        var = s2 / (v1 - v2/v1)   (for ddof=1),
    and supports the frequency-weight case by using
        var = s2 / (v1 - ddof)    when w is integer and v1 > 1.

    Parameters
    ----------
    a : np.array
        Input samples.
    w : np.array
        Associated sample weights.
        Integer -> frequency weights;
        Float -> reliability weights.
    axis : int
        Axis along which to compute variance.
    ddof : int, default=1
        Delta degrees of freedom.

    Returns
    -------
    var : float, np.array
        Unbiased variance.

    """
    mu = np.ma.filled(np.average(a, weights=w, axis=axis), np.nan)
    v1 = np.ma.filled(w.sum(axis=axis), 0.0)
    if np.issubdtype(w.dtype, np.integer):
        v2 = v1
    else:
        # ---- reliability weights branch ----
        v2 = np.ma.filled((w ** 2).sum(axis=axis), 0.0)
    s2 = np.ma.filled((w * (a - mu)**2).sum(axis=axis), 0.0)

    invalid = (v1 == 0) | (v1**2 - ddof*v2 == 0) | np.isnan(mu) | np.isnan(s2)
    nans = np.full_like(v1, np.nan, dtype=float)
    var = np.divide(v1 * s2, v1**2 - ddof*v2, out=nans.copy(), where=~invalid)
    return var if np.ndim(var) > 0 else np.float64(var)


def cov_unbiased(a, w, ddof=1, return_corr=False):
    """Compute the unbiased covariance from weighted samples.

    Parameters
    ----------
    a : np.array, shape (n_samples, n_features)
        Input samples in rows, features/parameters in columns.
    w : np.array, shape (n_samples,)
        Associated sample weights.
        Integer -> frequency weights;
        Float -> reliability weights.
    ddof : int, default=1
        Delta degrees of freedom.

    Returns
    -------
    cov : ndarray, shape (n_features, n_features)
        Unbiased covariance matrix.

    """
    a = np.atleast_2d(a)

    # If all weights are zero, mimic pandas' "all NaN" behaviour upstream.
    if w.sum() == 0:
        return np.full((a.shape[1], a.shape[1]), np.nan, dtype=float)

    # ---- fast path: no NaNs anywhere ----
    if not return_corr:
        if np.isfinite(a).all() and a.shape[0] >= 2 and a.shape[1] >= 2:
            if np.issubdtype(w.dtype, np.integer):
                return np.cov(a, rowvar=False, fweights=w, ddof=ddof)
            else:
                return np.cov(a, rowvar=False, aweights=w, ddof=ddof)

    M = np.isfinite(a)           # shape (n, p)
    A = np.where(M, a, 0.0)      # shape (n, p)
    WA = (w[:, None] * A).T      # shape (n, p)
    V1 = (w[:, None] * M).T @ M  # shape (p, p)
    S1 = WA @ M                  # shape (p, p)
    S2 = WA @ A                  # shape (p, p)

    if np.issubdtype(w.dtype, np.integer):
        # ---- frequency weights branch ----
        V2 = V1
    else:
        # ---- reliability weights branch ----
        V2 = ((w**2)[:, None] * M).T @ M  # shape (p, p)

    invalid = (V1 == 0) | (V1**2 - ddof * V2 == 0)
    cov = np.full_like(S2, np.nan, dtype=float)
    cov = np.divide(V1 * S2 - S1.T * S1, V1**2 - ddof * V2,
                    out=cov.copy(), where=~invalid)
    if not return_corr:
        return cov
    # pairwise variances on the SAME intersections
    S2i = (w[:, None] * A * A).T @ M
    nans = np.full_like(cov, np.nan, dtype=float)
    var_i = np.divide(V1 * S2i - S1**2, V1**2 - 1 * V2,
                      out=nans.copy(), where=~invalid)
    invalid = invalid | (var_i * var_i.T <= 0)
    corr = np.divide(cov, np.sqrt(var_i * var_i.T),
                     out=nans.copy(), where=~invalid)
    return corr


def skew_unbiased(a, w, axis=0):
    """Compute the unbiased skewness from weighted samples.

    Adapted from Lorenzo Rimoldini (2013):
    https://arxiv.org/pdf/1304.6564

    Parameters
    ----------
    a : np.array
        Input samples.
    w : np.array
        Associated sample weights.
        Integer -> frequency weights;
        Float -> reliability weights.
    axis : int
        Axis along which to compute kurtosis.

    Returns
    -------
    skew : float, np.array
        Unbiased skewness (`G1`).

    """
    mu = np.ma.filled(np.average(a, weights=w, axis=axis), np.nan)
    v1 = np.ma.filled(w.sum(axis=axis), 0.0)
    v2 = np.ma.filled((w**2).sum(axis=axis), 0.0)
    v3 = np.ma.filled((w**3).sum(axis=axis), 0.0)
    s2 = np.ma.filled((w * (a - mu)**2).sum(axis=axis), 0.0)
    s3 = np.ma.filled((w * (a - mu)**3).sum(axis=axis), 0.0)

    nans = np.full_like(v1, np.nan, dtype=float)

    # ---- frequency weights branch ----
    if np.issubdtype(w.dtype, np.integer):
        n = v1
        if np.all(n <= 2):
            return np.nan
        invalid = (n <= 2) | np.isnan(mu) | np.isnan(s2) | np.isnan(s3)

        # pandas convention: zero variance (s2==0) => skew = 0
        degenerate = (~invalid) & (s2 == 0)
        skew = np.divide(n * (n-1)**0.5 * s3, (n-2) * s2**1.5,
                         out=nans.copy(), where=(~invalid) & (~degenerate))
        skew = np.where(degenerate, 0.0, skew)
        return skew if np.ndim(skew) > 0 else np.float64(skew)

    # ---- reliability weights branch (Rimoldini) ----
    if a.shape[axis] <= 2:
        return np.nan
    invalid = ((v1 == 0) | (v1**2 - v2 == 0) | (v1**3 - 3*v1*v2 + 2*v3 == 0)
               | np.isnan(mu) | np.isnan(s2) | np.isnan(s3))
    k2 = np.divide(v1 * s2, v1**2 - v2, out=nans.copy(), where=~invalid)
    k3 = np.divide(v1**2 * s3, v1**3 - 3*v1*v2 + 2*v3,
                   out=nans.copy(), where=~invalid)

    # pandas convention: zero variance (k2==0) => skew = 0
    degenerate = (~invalid) & (k2 == 0)
    skew = np.divide(k3, k2**1.5, out=nans.copy(),
                     where=(~invalid) & (~degenerate))
    skew = np.where(degenerate, 0.0, skew)
    return skew if np.ndim(skew) > 0 else np.float64(skew)


def kurt_unbiased(a, w, axis=0):
    """Compute the unbiased kurtosis from weighted samples.

    Adapted from Lorenzo Rimoldini (2013):
    https://arxiv.org/pdf/1304.6564

    Parameters
    ----------
    a : np.array
        Input samples.
    w : np.array
        Associated sample weights.
        Integer -> frequency weights;
        Float -> reliability weights.
    axis : int
        Axis along which to compute kurtosis.

    Returns
    -------
    kurt : float, np.array
        Unbiased kurtosis (`G2`).

    """
    mu = np.ma.filled(np.average(a, weights=w, axis=axis), np.nan)
    v1 = np.ma.filled(w.sum(axis=axis), 0.0)
    v2 = np.ma.filled((w**2).sum(axis=axis), 0.0)
    v3 = np.ma.filled((w**3).sum(axis=axis), 0.0)
    v4 = np.ma.filled((w**4).sum(axis=axis), 0.0)
    s2 = np.ma.filled((w * (a - mu)**2).sum(axis=axis), 0.0)
    s4 = np.ma.filled((w * (a - mu)**4).sum(axis=axis), 0.0)

    nans = np.full_like(v1, np.nan, dtype=float)

    # ---- frequency weights branch ----
    if np.issubdtype(w.dtype, np.integer):
        n = v1
        if np.all(n <= 3):
            return np.nan
        invalid = (n <= 3) | np.isnan(mu) | np.isnan(s2) | np.isnan(s4)

        # pandas convention: zero variance (s2==0) => kurt = 0
        degenerate = (~invalid) & (s2 == 0)
        kurt = np.divide((n+1)*n*(n-1)*s4-3*(n-1)**2*s2**2, (n-2)*(n-3)*s2**2,
                         out=nans.copy(), where=(~invalid) & (~degenerate))
        kurt = np.where(degenerate, 0.0, kurt)
        return kurt if np.ndim(kurt) > 0 else np.float64(kurt)

    # ---- reliability weights branch (Rimoldini) ----
    if a.shape[axis] <= 3:
        return np.nan

    den2 = v1**2-v2
    den4 = (v1**2 - v2) * (v1**4 - 6*v1**2*v2 + 8*v1*v3 + 3*v2**2 - 6*v4)
    num41 = v1 * (v1**4 - 4*v1*v3 + 3*v2**2)
    num42 = 3 * (v1**4 - 2*v1**2*v2 + 4*v1*v3 - 3*v2**2)

    invalid = ((v1 == 0) | (den2 == 0) | (den4 == 0)
               | np.isnan(mu) | np.isnan(s2) | np.isnan(s4))
    k2 = np.divide(v1 * s2, den2, out=nans.copy(), where=~invalid)
    k4 = np.divide(num41*s4-num42*s2**2, den4, out=nans.copy(), where=~invalid)

    # pandas convention: zero variance (k2==0) => kurt = 0
    degenerate = (~invalid) & (k2 == 0)
    invalid = invalid | (k2 == 0)
    kurt = np.divide(k4, k2**2, out=nans.copy(), where=~invalid)
    kurt = np.where(degenerate, 0.0, kurt)
    return kurt if np.ndim(kurt) > 0 else np.float64(kurt)


def sample_cdf(samples, inverse=False, interpolation='linear'):
    """Sample the empirical cdf for a 1d array."""
    samples = np.sort(samples)
    ngaps = len(samples)-1
    gaps = np.random.dirichlet(np.ones(ngaps))
    cdf = np.array([0, *np.cumsum(gaps)])
    # The last element should be one (up to floating point errors) because
    # dirichlet sums to one. Set exactly to avoid interpolation errors.
    cdf[-1] = 1
    if inverse:
        return interp1d(cdf, samples, kind=interpolation)
    else:
        return interp1d(samples, cdf, kind=interpolation,
                        fill_value=(0, 1), bounds_error=False)


def credibility_interval(samples, weights=None, level=0.68, method="iso-pdf",
                         return_covariance=False, nsamples=12):
    """Compute the credibility interval of weighted samples.

    Based on linear interpolation of the cumulative density function, thus
    expect discretisation errors on the scale of distances between samples.

    https://github.com/Stefan-Heimersheim/fastCI#readme

    Parameters
    ----------
    samples : array
        Samples to compute the credibility interval of.
    weights : array, default=np.ones_like(samples)
        Weights corresponding to samples.
    level : float, default=0.68
        Credibility level (probability, <1).
    method : str, default='iso-pdf'
        Which definition of interval to use:

        * ``'iso-pdf'``: Calculate iso probability density interval with the
          same probability density at each end. Also known as
          waterline-interval or highest average posterior density interval.
          This is only accurate if the distribution is sufficiently uni-modal.
        * ``'lower-limit'``/``'upper-limit'``: Lower/upper limit. One-sided
          limits for which ``level`` fraction of the (equally weighted) samples
          lie above/below the limit.
        * ``'equal-tailed'``: Equal-tailed interval with the same fraction of
          (equally weighted) samples below and above the interval region.

    return_covariance: bool, default=False
        Return the covariance of the sampled limits, in addition to the mean
    nsamples : int, default=12
        Number of CDF samples to improve `mean` and `std` estimate.

    Returns
    -------
    limit(s) : float, array, or tuple of floats or arrays
        Returns the credibility interval boundari(es). By default,
        returns the mean over ``nsamples`` samples, which is either
        two numbers (``method='iso-pdf'``/``'equal-tailed'``) or one number
        (``method='lower-limit'``/``'upper-limit'``). If
        ``return_covariance=True``, returns a tuple (mean(s), covariance)
        where covariance is the covariance over the sampled limits.
    """
    if level >= 1:
        raise ValueError('level must be <1, got {0:.2f}'.format(level))
    if len(np.shape(samples)) != 1:
        raise ValueError('Support only 1D arrays for samples')
    if weights is not None and np.shape(samples) != np.shape(weights):
        raise ValueError('Shape of samples and weights differs')

    # Convert to numpy to unify indexing
    samples = np.array(samples.copy())
    if weights is None:
        weights = np.ones(len(samples))
    else:
        weights = np.array(weights.copy())

    # Convert samples to unit weight not the case
    if not np.all(np.logical_or(weights == 0, weights == 1)):
        # compress_weights with ncompress='equal' assures weights \in 0,1
        # Note that this must be done, we cannot handle weights != 1
        # see this discussion for details:
        # https://github.com/williamjameshandley/anesthetic/pull/188#issuecomment-1274980982
        weights = compress_weights(weights, ncompress='equal')

    indices = np.where(weights)[0]
    x = samples[indices]

    # Sample the confidence interval multiple times
    # to get errorbars on confidence interval boundaries
    ci_samples = []
    for i in range(nsamples):
        invCDF = sample_cdf(x, inverse=True)
        if method == 'iso-pdf':
            # Find smallest interval
            def distance(Y, level=level):
                return invCDF(Y+level)-invCDF(Y)
            res = minimize_scalar(distance, bounds=(0, 1-level),
                                  method="Bounded")
            ci_samples.append(np.array([invCDF(res.x),
                                        invCDF(res.x+level)]))
        elif method == 'lower-limit':
            # Get value from which we reach the desired level
            ci_samples.append(invCDF(1-level))
        elif method == 'upper-limit':
            # Get value to which we reach the desired level
            ci_samples.append(invCDF(level))
        elif method == 'equal-tailed':
            ci_samples.append(np.array([invCDF((1-level)/2),
                                        invCDF((1+level)/2)]))
        else:
            raise ValueError("Method '{0:}' unknown".format(method))
    ci_samples = np.array(ci_samples)
    if np.shape(ci_samples) == (nsamples, ):
        if return_covariance:
            return np.mean(ci_samples), np.cov(ci_samples)
        else:
            return np.mean(ci_samples)
    else:
        if return_covariance:
            return np.mean(ci_samples, axis=0), \
                   np.cov(ci_samples, rowvar=False)
        else:
            return np.mean(ci_samples, axis=0)


def mirror_1d(d, xmin=None, xmax=None):
    """If necessary apply reflecting boundary conditions."""
    if xmin is not None and xmax is not None:
        xmed = (xmin+xmax)/2
        return np.concatenate((2*xmin-d[d < xmed], d, 2*xmax-d[d >= xmed]))
    elif xmin is not None:
        return np.concatenate((2*xmin-d, d))
    elif xmax is not None:
        return np.concatenate((d, 2*xmax-d))
    else:
        return d


def mirror_2d(d_x_, d_y_, xmin=None, xmax=None, ymin=None, ymax=None):
    """If necessary apply reflecting boundary conditions."""
    d_x = d_x_.copy()
    d_y = d_y_.copy()

    if xmin is not None and xmax is not None:
        xmed = (xmin+xmax)/2
        d_y = np.concatenate((d_y[d_x < xmed], d_y, d_y[d_x >= xmed]))
        d_x = np.concatenate((2*xmin-d_x[d_x < xmed], d_x,
                              2*xmax-d_x[d_x >= xmed]))
    elif xmin is not None:
        d_y = np.concatenate((d_y, d_y))
        d_x = np.concatenate((2*xmin-d_x, d_x))
    elif xmax is not None:
        d_y = np.concatenate((d_y, d_y))
        d_x = np.concatenate((d_x, 2*xmax-d_x))

    if ymin is not None and ymax is not None:
        ymed = (ymin+ymax)/2
        d_x = np.concatenate((d_x[d_y < ymed], d_x, d_x[d_y >= ymed]))
        d_y = np.concatenate((2*ymin-d_y[d_y < ymed], d_y,
                              2*ymax-d_y[d_y >= ymed]))
    elif ymin is not None:
        d_x = np.concatenate((d_x, d_x))
        d_y = np.concatenate((2*ymin-d_y, d_y))
    elif ymax is not None:
        d_x = np.concatenate((d_x, d_x))
        d_y = np.concatenate((d_y, 2*ymax-d_y))

    return d_x, d_y


def nest_level(lst):
    """Calculate the nesting level of a list."""
    if not isinstance(lst, list):
        return 0
    if not lst:
        return 1
    return max(nest_level(item) for item in lst) + 1


def histogram(a, **kwargs):
    """Produce a histogram for path-based plotting.

    This is a cheap histogram. Necessary if one wants to update the histogram
    dynamically, and redrawing and filling is very expensive.

    This has the same arguments and keywords as :func:`numpy.histogram`, but is
    normalised to 1.
    """
    hist, bin_edges = np.histogram(a, **kwargs)
    xpath, ypath = np.empty((2, 4*len(hist)))
    ypath[0::4] = ypath[3::4] = 0
    ypath[1::4] = ypath[2::4] = hist
    xpath[0::4] = xpath[1::4] = bin_edges[:-1]
    xpath[2::4] = xpath[3::4] = bin_edges[1:]
    mx = max(ypath)
    if mx:
        ypath /= max(ypath)
    return xpath, ypath


def histogram_bin_edges(samples, weights, bins='fd', range=None, beta='equal'):
    """Compute a good number of bins dynamically from weighted samples.

    Parameters
    ----------
    samples : array_like
        Input data.

    weights : array-like
        Array of sample weights.

    bins : str, default='fd'
        String defining the rule used to automatically compute a good number
        of bins for the weighted samples:

            * 'fd'    : Freedman--Diaconis rule (modified for weighted data)
            * 'scott' : Scott's rule (modified for weighted data)
            * 'sqrt'  : Square root estimator (modified for weighted data)

    range : (float, float), optional
        The lower and upper range of the bins. If not provided, range is simply
        ``(a.min(), a.max())``. Values outside the range are ignored. The first
        element of the range must be less than or equal to the second.

    beta : float, default='equal'
        The value of beta>0 used to calculate the number of effective
        samples via :func:`neff`.

    Returns
    -------
    bin_edges : array of dtype float
        The edges to pass to :func:`numpy.histogram`.

    """
    if weights is None:
        weights = np.ones_like(samples)
    if range is None:
        minimum = np.min(samples)
        maximum = np.max(samples)
        data_range = maximum - minimum
    else:
        minimum = range[0]
        maximum = range[1]
        data_range = maximum - minimum
        data_mask = (samples >= minimum) & (samples <= maximum)
        samples = samples[data_mask]
        weights = weights[data_mask]
    w = weights / np.sum(weights)
    N_eff = neff(w=w, beta=beta)
    if bins == 'fd':  # Freedman--Diaconis rule
        q1, q3 = quantile(samples, [1/4, 3/4], w=w)
        bin_width = 2 * (q3 - q1) * N_eff**(-1/3)
    elif bins == 'scott':  # Scott's rule
        weighted_mean = np.average(samples, weights=w)
        weighted_var = np.average((samples - weighted_mean)**2, weights=w)
        weighted_std = np.sqrt(weighted_var)
        bin_width = (24 * np.pi**0.5 / N_eff)**(1/3) * weighted_std
    elif bins == 'sqrt':  # Square root estimator
        samples_eff, _ = sample_compression_1d(samples, w=w, ncompress=N_eff)
        data_range_eff = np.max(samples_eff) - np.min(samples_eff)
        bin_width = data_range_eff / np.sqrt(N_eff)
    nbins = int(np.ceil(data_range / bin_width))
    bin_edges = np.linspace(minimum, maximum, nbins+1)
    return bin_edges


def compute_nlive(death, birth):
    """Compute number of live points from birth and death contours.

    Parameters
    ----------
    death, birth : array-like
        list of birth and death contours

    Returns
    -------
    nlive : np.array
        number of live points at each contour
    """
    b = pandas.DataFrame(np.array(birth), columns=['logL'])
    d = pandas.DataFrame(np.array(death), columns=['logL'],
                         index=b.index + len(b))
    b['n'] = +1
    d['n'] = -1
    t = pandas.concat([b, d]).sort_values(['logL', 'n'], na_position='first')
    n = t.n.cumsum()
    return (n[d.index]+1).to_numpy()


def compute_insertion_indexes(death, birth):
    """Compute the live point insertion index for each point.

    For more detail, see `Fowlie et al. (2020)
    <https://arxiv.org/abs/2006.03371>`_

    Parameters
    ----------
    death, birth : array-like
        list of birth and death contours

    Returns
    -------
    indexes : np.array
        live point index at which each live point was inserted
    """
    indexes = np.zeros_like(birth, dtype=int)
    for i, (b, d) in enumerate(zip(birth, death)):
        i_live = (death > b) & (birth <= b)
        live = death[i_live]
        live.sort()
        indexes[i] = np.searchsorted(live, d)
    return indexes


def unique(a):
    """Find unique elements, retaining order."""
    b = []
    for x in a:
        if x not in b:
            b.append(x)
    return b


def iso_probability_contours(pdf, contours=[0.95, 0.68]):
    """Compute the iso-probability contour values."""
    if len(contours) > 1 and not np.all(contours[:-1] > contours[1:]):
        raise ValueError(
            "The kwargs `levels` and `contours` have to be ordered from "
            "outermost to innermost contour, i.e. in strictly descending "
            "order when referring to the enclosed probability mass, e.g. "
            "like the default [0.95, 0.68]. "
            "This breaking change in behaviour was introduced in version "
            "2.0.0-beta.10, in order to better match the ordering of other "
            "matplotlib kwargs."
        )
    contours = [1-p for p in contours]
    p = np.sort(np.array(pdf).flatten())
    m = np.cumsum(p)
    m /= m[-1]
    interp = interp1d([0]+list(m), [0]+list(p))
    c = list(interp(contours))+[max(p)]

    return c


def iso_probability_contours_from_samples(pdf, contours=[0.95, 0.68],
                                          weights=None):
    """Compute the iso-probability contour values."""
    if len(contours) > 1 and not np.all(contours[:-1] > contours[1:]):
        raise ValueError(
            "The kwargs `levels` and `contours` have to be ordered from "
            "outermost to innermost contour, i.e. in strictly descending "
            "order when referring to the enclosed probability mass, e.g. "
            "like the default [0.95, 0.68]. "
            "This breaking change in behaviour was introduced in version "
            "2.0.0-beta.10, in order to better match the ordering of other "
            "matplotlib kwargs."
        )
    if weights is None:
        weights = np.ones_like(pdf)
    contours = [1-p for p in contours]
    i = np.argsort(pdf)
    m = np.cumsum(weights[i])
    m /= m[-1]
    interp = interp1d([0]+list(m), [0]+list(pdf[i]))
    c = list(interp(contours))+[max(pdf)]

    return c


def scaled_triangulation(x, y, cov):
    """Triangulation scaled by a covariance matrix.

    Parameters
    ----------
    x, y : array-like
        x and y coordinates of samples

    cov : array-like, 2d
        Covariance matrix for scaling

    Returns
    -------
    :class:`matplotlib.tri.Triangulation`
        Triangulation with the appropriate scaling
    """
    L = np.linalg.cholesky(cov)
    Linv = np.linalg.inv(L)
    x_, y_ = Linv.dot([x, y])
    tri = Triangulation(x_, y_)
    return Triangulation(x, y, tri.triangles)


def triangular_sample_compression_2d(x, y, cov, w=None, n=1000):
    """Histogram a 2D set of weighted samples via triangulation.

    This defines bins via a triangulation of the subsamples and sums weights
    within triangles surrounding each point

    Parameters
    ----------
    x, y : array-like
        x and y coordinates of samples for compressing

    cov : array-like, 2d
        Covariance matrix for scaling

    w : :class:`pandas.Series`, optional
        weights of samples

    n : int, default=1000
        number of samples returned.

    Returns
    -------
    tri :
        :class:`matplotlib.tri.Triangulation` with an appropriate scaling

    w : array-like
        Compressed samples and weights
    """
    # Pre-process samples to not be affected by non-standard indexing
    # Details: https://github.com/handley-lab/anesthetic/issues/189
    x = np.array(x)
    y = np.array(y)

    x = pandas.Series(x)
    if w is None:
        w = pandas.Series(index=x.index, data=np.ones_like(x))

    if n is False:
        n = len(x)
    elif n is True or isinstance(n, str):
        if n is True:
            n = 'entropy'
        n = int(neff(w, beta=n))

    # Select samples for triangulation
    if (w != 0).sum() <= n:
        i = x.index
    else:
        i = np.random.choice(x.index, size=n, replace=False, p=w/w.sum())

    # Generate triangulation
    tri = scaled_triangulation(x[i], y[i], cov)

    # For each point find corresponding triangles
    trifinder = tri.get_trifinder()
    j = trifinder(x, y)
    k = tri.triangles[j[j != -1]]

    # Compute mass in each triangle, and add it to each corner
    w_ = np.zeros(len(i))
    for i in range(3):
        np.add.at(w_, k[:, i], w[j != -1]/3)

    return tri, w_


def sample_compression_1d(x, w=None, ncompress=True):
    """Histogram a 1D set of weighted samples via subsampling.

    This compresses the number of samples, combining weights.

    Parameters
    ----------
    x : array-like
        x coordinate of samples for compressing

    w : :class:`pandas.Series`, optional
        weights of samples

    ncompress : int, default=True
        Degree of compression.

        * If ``int``: number of samples returned.
        * If ``True``: compresses to the channel capacity
          (same as ``ncompress='entropy'``).
        * If ``False``: no compression.
        * If ``str``: determine number from the Huggins-Roy family of effective
          samples in :func:`neff` with ``beta=ncompress``.

    Returns
    -------
    x, w: array-like
        Compressed samples and weights
    """
    if ncompress is False:
        return x, w
    elif ncompress is True or isinstance(ncompress, str):
        if ncompress is True:
            ncompress = 'entropy'
        ncompress = neff(w, beta=ncompress)
    x = np.array(x)
    if w is None:
        w = np.ones_like(x)
    w = np.array(w)

    # Select inner samples for triangulation
    if len(x) > ncompress:
        x_ = np.random.choice(x, size=int(ncompress), replace=False)
    else:
        x_ = x.copy()
    x_.sort()

    # Compress mass onto these subsamples
    centers = (x_[1:] + x_[:-1])/2
    j = np.digitize(x, centers)
    w_ = np.zeros_like(x_)
    np.add.at(w_, j, w)

    return x_, w_


def is_int(x):
    """Test whether x is an integer."""
    return isinstance(x, int) or isinstance(x, np.integer)


def match_contour_to_contourf(contours, vmin, vmax):
    """Get needed `vmin, vmax` to match `contour` colors to `contourf` colors.

    `contourf` uses the arithmetic mean of contour levels to assign colors,
    whereas `contour` uses the contour level directly. To get the same colors
    for `contour` lines as for `contourf` faces, we need some fiddly algebra.
    """
    if len(contours) <= 2:
        vmin = 2 * vmin - vmax
        return vmin, vmax
    else:
        c0 = contours[0]
        c1 = contours[1]
        ce = contours[-2]
        denom = vmax + ce - c1 - c0
        vmin = +(c0 * vmax - c1 * ce + 2 * vmin * (ce - c0)) / denom
        vmax = -(c0 * vmax + c1 * ce - 2 * vmax * ce) / denom
        return vmin, vmax


def insertion_p_value(indexes, nlive, batch=0):
    """Compute the p-value from insertion indexes, assuming constant nlive.

    Note that this function doesn't use :func:`scipy.stats.kstest` as the
    latter assumes continuous distributions.

    For more detail, see `Fowlie et al. (2020)
    <https://arxiv.org/abs/2006.03371>`_

    For a rolling test, you should provide the optional parameter ``batch!=0``.
    In this case the test computes the p-value on consecutive batches of size
    ``nlive * batch``, selects the smallest one and adjusts for multiple
    comparisons using a Bonferroni correction.

    Parameters
    ----------
    indexes : array-like
        list of insertion indexes, sorted by death contour

    nlive : int
        number of live points

    batch : float
        batch size in units of nlive for a rolling p-value

    Returns
    -------
    ks_result : dict
        Kolmogorov-Smirnov test results:

            * ``D``: Kolmogorov-Smirnov statistic
            * ``sample_size``: sample size
            * ``p-value``: p-value

            if ``batch != 0``:

            * ``iterations``: bounds of batch with minimum p-value
            * ``nbatches``: the number of batches in total
            * ``uncorrected p-value``: p-value without Bonferroni correction
    """
    if batch == 0:
        bins = np.arange(-0.5, nlive + 0.5, 1.)
        empirical_pmf = np.histogram(
            np.array(indexes),
            bins=bins,
            density=True,
        )[0]
        empirical_cmf = np.cumsum(empirical_pmf)
        uniform_cmf = np.arange(1., nlive + 1., 1.) / nlive

        D = abs(empirical_cmf - uniform_cmf).max()
        sample_size = len(indexes)
        K = D * np.sqrt(sample_size)

        ks_result = {}
        ks_result["D"] = D
        ks_result["sample_size"] = sample_size
        ks_result["p-value"] = kstwobign.sf(K)
        return ks_result
    else:
        batch = int(batch * nlive)
        batches = [
            np.array(indexes)[i:i + batch]
            for i in range(0, len(indexes), batch)
        ]
        ks_results = [insertion_p_value(c, nlive) for c in batches]
        ks_result = min(ks_results, key=lambda t: t["p-value"])
        index = ks_results.index(ks_result)

        ks_result["iterations"] = (index * batch, (index + 1) * batch)
        ks_result["nbatches"] = n = len(batches)
        ks_result["uncorrected p-value"] = p = ks_result["p-value"]
        ks_result["p-value"] = 1. - (1. - p)**n if p != 1 else p * n
        return ks_result


@contextlib.contextmanager
def temporary_seed(seed):
    """Context for temporarily setting a numpy seed."""
    state = np.random.get_state()
    np.random.seed(seed)
    try:
        yield
    finally:
        np.random.set_state(state)


def adjust_docstrings(obj, pattern, repl, *args, **kwargs):
    """Adjust the docstrings of a class using regular expressions.

    After the first argument, the remaining arguments are identical to re.sub.

    Parameters
    ----------
    cls : class
        class to adjust

    pattern : str
        regular expression pattern

    repl : str
        replacement string
    """
    if inspect.isclass(obj):
        for key, val in obj.__dict__.items():
            doc = inspect.getdoc(val)
            if doc is not None:
                i = doc.find("See Also")
                j = doc.find("Examples")
                if i != -1:
                    doc = doc[:i] if j == -1 else doc[:i] + doc[j:]
                newdoc = re.sub(pattern, repl, doc, *args, **kwargs)
                try:
                    obj.__dict__[key].__doc__ = newdoc
                except AttributeError:
                    pass
    else:
        doc = inspect.getdoc(obj)
        if doc is not None:
            i = doc.find("See Also")
            j = doc.find("Examples")
            if i != -1:
                doc = doc[:i] if j == -1 else doc[:i] + doc[j:]
            newdoc = re.sub(pattern, repl, doc, *args, **kwargs)
            obj.__doc__ = newdoc



"""Read NestedSamples from UltraNest results."""
import os
import json
from anesthetic.samples import NestedSamples


def read_ultranest(root, *args, **kwargs):
    """Read UltraNest files.

    Parameters
    ----------
    root : str
        root name for reading files in UltraNest format, i.e. the files
        ``<root>/info/results.json`` and ``<root>/results/points.hdf5``.

    """
    with open(os.path.join(root, 'info', 'results.json')) as infofile:
        labels = json.load(infofile)['paramnames']
    num_params = len(labels)

    filepath = os.path.join(root, 'results', 'points.hdf5')
    try:
        import h5py
    except ImportError:
        raise ImportError('h5py is required to read UltraNest results')
    with h5py.File(filepath, 'r') as fileobj:
        points = fileobj['points']
        _, ncols = points.shape
        x_dim = ncols - 3 - num_params
        logL_birth = points[:, 0]
        logL = points[:, 1]
        samples = points[:, 3+x_dim:3+x_dim+num_params]

    kwargs['label'] = kwargs.get('label', os.path.basename(root))
    columns = kwargs.pop('columns', labels)
    labels = kwargs.pop('labels', labels)
    data = samples

    return NestedSamples(data=data, logL=logL, logL_birth=logL_birth,
                         columns=columns, labels=labels, *args, **kwargs)



"""Pandas DataFrame and Series with weighted samples."""

import warnings
from inspect import signature
import numpy as np
from pandas import Series, DataFrame, concat, MultiIndex
from pandas.core.groupby import GroupBy, SeriesGroupBy, DataFrameGroupBy, ops
from pandas._libs import lib
from pandas._libs.lib import no_default
from pandas.util._exceptions import find_stack_level
from pandas.util import hash_pandas_object
from anesthetic.utils import (compress_weights, neff, quantile,
                              temporary_seed, adjust_docstrings,
                              var_unbiased, cov_unbiased, skew_unbiased,
                              kurt_unbiased, credibility_interval)
from pandas.core.accessor import CachedAccessor
from anesthetic.plotting import PlotAccessor
import pandas as pd


def read_csv(filename, *args, **kwargs):
    """Read a CSV file into a ``WeightedDataFrame``."""
    df = pd.read_csv(filename, index_col=[0, 1], header=[0, 1],
                     *args, **kwargs)
    wdf = WeightedDataFrame(df)
    if wdf.isweighted(0) and wdf.isweighted(1):
        wdf.set_weights(wdf.get_weights(axis=1).astype(float),
                        axis=1, inplace=True)
        return wdf
    df = pd.read_csv(filename, index_col=[0, 1], *args, **kwargs)
    wdf = WeightedDataFrame(df)
    if wdf.isweighted(0):
        return wdf
    df = pd.read_csv(filename, index_col=0, header=[0, 1], *args, **kwargs)
    wdf = WeightedDataFrame(df)
    if wdf.isweighted(1):
        wdf.set_weights(wdf.get_weights(axis=1).astype(float),
                        axis=1, inplace=True)
        return wdf
    df = pd.read_csv(filename, index_col=0, *args, **kwargs)
    return WeightedDataFrame(df)


class WeightedGroupBy(GroupBy):
    """Weighted version of ``pandas.core.groupby.GroupBy``."""

    _grouper: ops.BaseGrouper

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _add_weights(self, name, *args, **kwargs):
        result = self.agg(lambda df: getattr(self.obj._constructor(df), name)
                          (*args, **kwargs)).set_weights(self.get_weights())
        return result.__finalize__(self.obj, method="groupby")

    def mean(self, **kwargs):  # noqa: D102
        return self._add_weights("mean", **kwargs)

    def std(self, **kwargs):  # noqa: D102
        return self._add_weights("std", **kwargs)

    def median(self, **kwargs):  # noqa: D102
        return self._add_weights("median", **kwargs)

    def var(self, **kwargs):  # noqa: D102
        return self._add_weights("var", **kwargs)

    def kurt(self, **kwargs):  # noqa: D102
        return self._add_weights("kurt", **kwargs)

    def kurtosis(self, **kwargs):  # noqa: D102
        return self._add_weights("kurtosis", **kwargs)

    def sem(self, **kwargs):  # noqa: D102
        return self._add_weights("sem", **kwargs)

    def skew(self, **kwargs):  # noqa: D102
        return self._add_weights("skew", **kwargs)

    def quantile(self, *args, **kwargs):  # noqa: D102
        return self._add_weights("quantile", *args, **kwargs)

    def get_weights(self):
        """Return the weights of the grouped samples."""
        return self.agg(lambda df: df.get_weights().sum())

    def _op_via_apply(self, name, *args, **kwargs):
        result = super()._op_via_apply(name, *args, **kwargs)
        try:
            index = result.index.get_level_values(self.keys)
            weights = self.get_weights()[index]
        except KeyError:
            weights = self.get_weights()
        return result.set_weights(weights, level=1)


class WeightedSeriesGroupBy(WeightedGroupBy, SeriesGroupBy):
    """Weighted version of ``pandas.core.groupby.SeriesGroupBy``."""

    def sample(self, *args, **kwargs):  # noqa: D102
        return super().sample(weights=self.obj.get_weights(), *args, **kwargs)

    def cov(self, *args, **kwargs):  # noqa: D102
        return self._op_via_apply("cov", *args, **kwargs)


class WeightedDataFrameGroupBy(WeightedGroupBy, DataFrameGroupBy):
    """Weighted version of ``pandas.core.groupby.DataFrameGroupBy``."""

    def get_weights(self):
        """Return the weights of the grouped samples."""
        return super().get_weights().min(axis=1)

    def _gotitem(self, key, ndim: int, subset=None):  # pragma: no cover
        if ndim == 2:
            if subset is None:
                subset = self.obj
            return WeightedDataFrameGroupBy(
                subset,
                self._grouper,
                level=self.level,
                grouper=self._grouper,
                exclusions=self.exclusions,
                selection=key,
                as_index=self.as_index,
                sort=self.sort,
                group_keys=self.group_keys,
                observed=self.observed,
                dropna=self.dropna,
            )
        elif ndim == 1:
            if subset is None:
                subset = self.obj[key]
            return WeightedSeriesGroupBy(
                subset,
                level=self.level,
                grouper=self._grouper,
                selection=key,
                sort=self.sort,
                group_keys=self.group_keys,
                observed=self.observed,
                dropna=self.dropna,
            )

        raise AssertionError("invalid ndim for _gotitem")

    def sample(self, *args, **kwargs):  # noqa: D102
        return super().sample(weights=self.obj.get_weights(), *args, **kwargs)

    def cov(self, *args, **kwargs):  # noqa: D102
        return self._op_via_apply("cov", *args, **kwargs)


class _WeightedObject(object):
    """Common methods for `WeightedSeries` and `WeightedDataFrame`.

    :meta public:
    """

    def __init__(self, *args, **kwargs):
        weights = kwargs.pop('weights', None)
        super().__init__(*args, **kwargs)
        if weights is not None:
            self.set_weights(weights, inplace=True)

    plot = CachedAccessor("plot", PlotAccessor)
    """:meta private:"""

    def isweighted(self, axis=0):
        """Determine if weights are actually present."""
        return 'weights' in self._get_axis(axis).names

    def get_weights(self, axis=0):
        """Retrieve sample weights from an axis."""
        if self.isweighted(axis):
            return self._get_axis(axis).get_level_values('weights').to_numpy()
        else:
            return np.ones_like(self._get_axis(axis), dtype=int)

    def drop_weights(self, axis=0):
        """Drop weights."""
        if self.isweighted(axis):
            return self.droplevel('weights', axis)
        return self.copy().__finalize__(self, "drop_weights")

    def set_weights(self, weights, axis=0, inplace=False, level=None):
        """Set sample weights along an axis.

        Parameters
        ----------
        weights : 1d array-like
            The sample weights to put in an index.

        axis : int (0,1), default=0
            Whether to put weights in an index or column.

        inplace : bool, default=False
            Whether to operate inplace, or return a new array.

        level : int
            Which level in the index to insert before.
            Defaults to inserting at back

        """
        if inplace:
            result = self
        else:
            result = self.copy()

        if weights is None:
            if result.isweighted(axis=axis):
                result = result.drop_weights(axis)
        else:
            names = [n for n in result._get_axis(axis).names if n != 'weights']
            index = [result._get_axis(axis).get_level_values(n) for n in names]
            if level is None:
                if result.isweighted(axis):
                    level = result._get_axis(axis).names.index('weights')
                else:
                    level = len(index)
            index.insert(level, weights)
            names.insert(level, 'weights')

            index = MultiIndex.from_arrays(index, names=names)
            result = result.set_axis(index, axis=axis)

        if inplace:
            self._update_inplace(result)
        else:
            return result.__finalize__(self, "set_weights")

    def _rand(self, axis=0):
        """Random number for consistent compression."""
        seed = hash_pandas_object(self._get_axis(axis)).sum() % 2**32
        with temporary_seed(int(seed)):
            return np.random.rand(self.shape[axis])

    def _weighted_stat(self, func, axis=0, skipna=True, **kwargs):
        """Compute weighted statistics using common pattern."""
        if not self.isweighted(axis):
            # Get the calling method name automatically
            method_name = func.__name__

            # Check if the method exists in pandas DataFrame
            if hasattr(super(), method_name):
                return getattr(super(), method_name)(axis=axis, skipna=skipna,
                                                     **kwargs)

        if self.get_weights(axis).sum() == 0:
            return self._constructor_sliced(np.nan,
                                            index=self._get_axis(1-axis))

        na = self.isna() & skipna
        weights = np.broadcast_to(
            self.get_weights(axis)[..., None] if axis == 0 else
            self.get_weights(axis)[None, ...],
            self.shape
        )
        if skipna:
            weights = np.ma.array(weights, mask=na)
        result = np.ma.filled(
            func(self, na=na, w=weights, axis=axis, skipna=skipna, **kwargs),
            np.nan
        )
        return self._constructor_sliced(result, index=self._get_axis(1-axis))

    def reset_index(self, level=None, drop=False, inplace=False,
                    *args, **kwargs):
        """Reset the index, retaining weights."""
        weights = self.get_weights()
        answer = super().reset_index(level=level, drop=drop,
                                     inplace=False, *args, **kwargs)
        answer.set_weights(weights, inplace=True)
        if inplace:
            self._update_inplace(answer)
        else:
            return answer.__finalize__(self, "reset_index")

    def neff(self, axis=0, beta=1):
        """Effective number of samples."""
        if self.isweighted(axis):
            return neff(self.get_weights(axis), beta=beta)
        else:
            return self.shape[axis]


class WeightedSeries(_WeightedObject, Series):
    """Weighted version of :class:`pandas.Series`."""

    def mean(self, skipna=True):  # noqa: D102
        na = self.isna() & skipna
        weights = self.get_weights()
        if skipna:
            weights = np.ma.array(self.get_weights(), mask=na)
        if weights.sum() == 0 or skipna and na.all():
            return np.nan
        return np.average(np.ma.array(self, mask=na), weights=weights)

    def std(self, skipna=True, **kwargs):  # noqa: D102
        return np.sqrt(self.var(skipna=skipna, **kwargs))

    def kurtosis(self, **kwargs):  # noqa: D102
        return self.kurt(**kwargs)

    def median(self, **kwargs):  # noqa: D102
        if self.get_weights().sum() == 0:
            return np.nan
        return self.quantile(**kwargs)

    def var(self, skipna=True, **kwargs):  # noqa: D102
        na = self.isna() & skipna
        w = np.ma.array(self.get_weights(), mask=na)
        if na.all() or self.isna().any() and not skipna or w.sum() == 0:
            return np.float64(np.nan)
        return var_unbiased(np.ma.array(self, mask=na), w, **kwargs)

    def cov(self, other, ddof=1, **kwargs):  # noqa: D102
        w = self.get_weights()
        x, y = self.align(other, join="inner")
        if len(x) == 0:
            return np.nan
        valid = x.notna() & y.notna()
        x = x[valid]
        y = y[valid]
        w = w[valid]
        if len(x) == 0 or w.sum() == 0:
            return np.nan
        X = np.column_stack((x.to_numpy(dtype=float), y.to_numpy(dtype=float)))
        return cov_unbiased(X, w, ddof=ddof)[0, 1]

    def corr(self, other, **kwargs):  # noqa: D102
        if not self.isweighted():
            return super().corr(other, **kwargs)
        if self.isna().all():
            return np.nan
        norm = (self.std(skipna=True, ddof=1) *
                other.std(skipna=True, ddof=1))
        if norm == 0:
            return np.float64(np.nan)
        return self.cov(other, ddof=1) / norm

    def kurt(self, skipna=True, **kwargs):  # noqa: D102
        if self.isna().all() or self.isna().any() and not skipna:
            return np.nan if skipna or self.size == 1 else np.float64(np.nan)
        na = self.isna() & skipna
        w = np.ma.array(self.get_weights(), mask=na)
        return kurt_unbiased(np.ma.array(self, mask=na), w)

    def skew(self, skipna=True, **kwargs):  # noqa: D102
        if self.isna().all() or self.isna().any() and not skipna:
            return np.nan if skipna or self.size == 1 else np.float64(np.nan)
        na = self.isna() & skipna
        w = np.ma.array(self.get_weights(), mask=na)
        return skew_unbiased(np.ma.array(self, mask=na), w)

    def sem(self, skipna=True, ddof=1, **kwargs):  # noqa: D102
        na = self.isna() & skipna
        w = np.ma.array(self.get_weights(), mask=na)
        V1 = w.sum()
        if np.issubdtype(w.dtype, np.integer) and V1 > 1:
            # frequency weights
            n = np.ma.filled(V1, np.nan)
        else:
            # reliability weights
            n = np.ma.filled(V1**2 / (w**2).sum(), np.nan)
        return np.sqrt(self.var(skipna=skipna, ddof=ddof, **kwargs) / n)

    def quantile(self, q=0.5, interpolation='linear'):  # noqa: D102
        if self.get_weights().sum() == 0:
            return np.nan
        return quantile(self.to_numpy(), q, self.get_weights(), interpolation)

    def compress(self, ncompress=True, weighted=False):
        """Reduce the number of samples by discarding low-weights.

        Parameters
        ----------
        ncompress : int, float, str, default=True
            Degree of compression.

            * If ``True`` (default): reduce to the channel capacity
              (theoretical optimum compression), equivalent to
              ``ncompress='entropy'``.
            * If ``> 0``: desired number of samples after compression.
            * If ``<= 0``: compress so that all remaining weights are unity.
            * If ``str``: determine number from the Huggins-Roy family of
              effective samples in :func:`anesthetic.utils.neff`
              with ``beta=ncompress``.

        weighted : bool, default=False
            If False (default), return an unweighted object with potentially
            repeated samples.
            If True, return a weighted object with non-zero compressed weights.

        """
        if (not self.isweighted() and isinstance(ncompress, (bool, str))
                or ncompress is False):
            return self
        w = compress_weights(self.get_weights(), self._rand(), ncompress)
        if weighted:
            mask = w > 0
            return self.drop_weights()[mask].set_weights(w[mask])
        else:
            return self.drop_weights().repeat(w)

    def sample(self, *args, **kwargs):  # noqa: D102
        return super().sample(weights=self.get_weights(), *args, **kwargs)

    def credibility_interval(self, level=0.68, method="iso-pdf",
                             return_covariance=False, nsamples=12):
        """Compute the credibility interval of the weighted samples.

        Based on linear interpolation of the cumulative density function, thus
        expect discretisation errors on the scale of distances between samples.

        https://github.com/Stefan-Heimersheim/fastCI#readme

        Parameters
        ----------
        level : float, default=0.68
            Credibility level (probability, <1).
        method : str, default='iso-pdf'
            Which definition of interval to use:

            * ``'iso-pdf'``: Calculate iso probability density interval with
              the same probability density at each end. Also known as
              waterline-interval or highest average posterior density interval.
              This is only accurate if the distribution is sufficiently
              uni-modal.
            * ``'lower-limit'``/``'upper-limit'``: Lower/upper limit. One-sided
              limits for which ``level`` fraction of the (equally weighted)
              samples lie above/below the limit.
            * ``'equal-tailed'``: Equal-tailed interval with the same fraction
              of (equally weighted) samples below and above the interval
              region.

        return_covariance: bool, default=False
            Return the covariance of the sampled limits, in addition to the
            mean
        nsamples : int, default=12
            Number of CDF samples to improve `mean` and `std` estimate.

        Returns
        -------
        limit(s) : float, array, or tuple of floats or arrays
            Returns the credibility interval boundaries of the Series.
            By default, returns the mean over ``nsamples`` samples, which is
            either two numbers (``method='iso-pdf'``/``'equal-tailed'``) or
            one number (``method='lower-limit'``/``'upper-limit'``). If
            ``return_covariance=True``, returns a tuple (mean(s), covariance)
            where covariance is the covariance over the sampled limits.
        """
        return credibility_interval(self, weights=self.get_weights(),
                                    level=level, method=method,
                                    return_covariance=return_covariance,
                                    nsamples=nsamples)

    @property
    def _constructor(self):
        return WeightedSeries

    @property
    def _constructor_expanddim(self):
        return WeightedDataFrame

    def groupby(
        self,
        by=None,
        axis=0,
        level=None,
        as_index=True,
        sort=True,
        group_keys=True,
        observed=False,
        dropna=True,
    ):  # pragma: no cover  # noqa: D102
        if level is None and by is None:
            raise TypeError("You have to supply one of 'by' and 'level'")
        if not as_index:
            raise TypeError("as_index=False only valid with DataFrame")
        axis = self._get_axis_number(axis)

        return WeightedSeriesGroupBy(
            obj=self,
            keys=by,
            level=level,
            as_index=as_index,
            sort=sort,
            group_keys=group_keys,
            observed=observed,
            dropna=dropna,
        )


class WeightedDataFrame(_WeightedObject, DataFrame):
    """Weighted version of :class:`pandas.DataFrame`."""

    def mean(self, axis=0, skipna=True, **kwargs):  # noqa: D102
        def mean(data, na, w, axis, skipna, **kwargs):
            if skipna:
                data = np.ma.array(data, mask=na)
            return np.average(data, weights=w, axis=axis)

        return self._weighted_stat(mean, axis, skipna, **kwargs)

    def std(self, axis=0, skipna=True, **kwargs):  # noqa: D102
        return np.sqrt(self.var(axis=axis, skipna=skipna, **kwargs))

    def kurtosis(self, **kwargs):  # noqa: D102
        return self.kurt(**kwargs)

    def median(self, **kwargs):  # noqa: D102
        return self.quantile(**kwargs)

    def var(self, axis=0, skipna=True, **kwargs):  # noqa: D102
        def var(data, na, w, axis, skipna, **kwargs):
            if skipna:
                data = np.ma.array(data, mask=na)
            return var_unbiased(data, w, axis=axis, ddof=kwargs.pop('ddof', 1))
        return self._weighted_stat(var, axis, skipna, **kwargs)

    def cov(self, ddof=1, **kwargs):  # noqa: D102
        if kwargs:
            raise TypeError(f"WeightedDataFrame.cov() got unexpected keyword "
                            f"arguments {kwargs}")

        if not self.isweighted():
            return super().cov(ddof=ddof)

        cov = cov_unbiased(self, self.get_weights(), ddof=ddof)
        return self._constructor(cov, index=self.columns, columns=self.columns)

    def corr(self, **kwargs):  # noqa: D102
        if not self.isweighted():
            return super().corr(**kwargs)
        corr = cov_unbiased(self, self.get_weights(), ddof=1, return_corr=True)
        return self._constructor(corr, index=self.columns,
                                 columns=self.columns)

    def corrwith(self, other, axis=0, drop=False, **kwargs):  # noqa: D102
        axis = self._get_axis_number(axis)
        if not self.isweighted(axis):
            return super().corrwith(other, drop=drop, axis=axis, **kwargs)
        else:
            if isinstance(other, Series):
                answer = self.apply(lambda x: other.corr(x, **kwargs),
                                    axis=axis)
                return self._constructor_sliced(answer)

            left, right = self.align(other, join="inner")

            if axis == 1:
                left = left.T
                right = right.T

            weights = left.index.to_frame()['weights']
            weights, _ = weights.align(right, join="inner")

            # mask missing values
            left = left + right * 0
            right = right + left * 0

            # demeaned data
            ldem = left - left.mean()
            rdem = right - right.mean()

            ddof = kwargs.pop('ddof', 0)
            num = (ldem * rdem * weights.to_numpy()[:, None]).sum()
            dom = weights.sum() * left.std(ddof=ddof) * right.std(ddof=ddof)

            correl = num / dom

            if not drop:
                # Find non-matching labels along the given axis
                result_index = self._get_axis(1-axis).union(
                    other._get_axis(1-axis)
                )
                idx_diff = result_index.difference(correl.index)

                if len(idx_diff) > 0:
                    correl = concat([
                        correl,
                        Series([np.nan] * len(idx_diff), index=idx_diff)
                    ])

            return self._constructor_sliced(correl)

    def kurt(self, axis=0, skipna=True, **kwargs):  # noqa: D102
        def kurt(data, na, w, axis, skipna, **kwargs):
            if skipna:
                data = np.ma.array(data, mask=na)
            return kurt_unbiased(data, w, axis=axis)
        return self._weighted_stat(kurt, axis, skipna, **kwargs)

    def skew(self, axis=0, skipna=True, **kwargs):  # noqa: D102
        def skew(data, na, w, axis, skipna, **kwargs):
            if skipna:
                data = np.ma.array(data, mask=na)
            return skew_unbiased(data, w, axis=axis)
        return self._weighted_stat(skew, axis, skipna, **kwargs)

    def sem(self, axis=0, skipna=True, **kwargs):  # noqa: D102
        def sem(data, na, w, axis, skipna, **kwargs):
            V1 = w.sum(axis=axis)
            if np.issubdtype(w.dtype, np.integer) and np.all(V1 > 1):
                # frequency weights
                n = V1
            else:
                # reliability weights
                n = V1**2 / (w**2).sum(axis=axis)
            return np.sqrt(self.var(axis=axis, skipna=skipna, **kwargs) / n)
        return self._weighted_stat(sem, axis, skipna, **kwargs)

    def quantile(self, q=0.5, axis=0, numeric_only=None,
                 interpolation='linear', method=None):  # noqa: D102
        if self.isweighted(axis):
            if numeric_only is not None or method is not None:
                raise NotImplementedError(
                    "`numeric_only` and `method` kwargs not implemented for "
                    "`WeightedSeries` and `WeightedDataFrame`."
                )
            data = np.array([c.quantile(q=q, interpolation=interpolation)
                             for _, c in self.items()])
            if np.isscalar(q):
                return self._constructor_sliced(data,
                                                index=self._get_axis(1-axis))
            else:
                return self._constructor(data.T, index=q,
                                         columns=self._get_axis(1-axis))
        else:
            if numeric_only is None:
                numeric_only = True
            if method is None:
                method = 'single'
            return super().quantile(q=q, axis=axis, numeric_only=numeric_only,
                                    interpolation=interpolation, method=method)

    def compress(self, ncompress=True, axis=0, weighted=False):
        """Reduce the number of samples by discarding low-weights.

        Parameters
        ----------
        ncompress : int, float, str, default=True
            Degree of compression.

            * If ``True`` (default): reduce to the channel capacity
              (theoretical optimum compression), equivalent to
              ``ncompress='entropy'``.
            * If ``> 0``: desired number of samples after compression.
            * If ``<= 0``: compress so that all remaining weights are unity.
            * If ``str``: determine number from the Huggins-Roy family of
              effective samples in :func:`anesthetic.utils.neff`
              with ``beta=ncompress``.

        weighted : bool, default=False
            If False (default), return an unweighted object with potentially
            repeated samples.
            If True, return a weighted object with non-zero compressed weights.

        """
        if (not self.isweighted(axis) and isinstance(ncompress, (bool, str))
                or ncompress is False):
            return self
        w = compress_weights(self.get_weights(axis), self._rand(axis),
                             ncompress)
        df = self.drop_weights(axis)
        if weighted:
            indices = np.flatnonzero(w > 0)
            df = df.take(indices, axis=axis)
            return df.set_weights(w[indices], axis=axis)
        else:
            indices = np.repeat(np.arange(df.shape[axis]), w)
            return df.take(indices, axis=axis)

    def sample(self, *args, **kwargs):  # noqa: D102
        sig = signature(DataFrame.sample)
        axis = sig.bind(self, *args, **kwargs).arguments.get('axis', 0)
        if self.isweighted(axis):
            return super().sample(weights=self.get_weights(axis),
                                  *args, **kwargs)
        else:
            return super().sample(*args, **kwargs)

    def credibility_interval(self, level=0.68, method="iso-pdf",
                             return_covariance=False, nsamples=12):
        """Compute the credibility interval of the weighted samples.

        Based on linear interpolation of the cumulative density function, thus
        expect discretisation errors on the scale of distances between samples.

        https://github.com/Stefan-Heimersheim/fastCI#readme

        Parameters
        ----------
        level : float, default=0.68
            Credibility level (probability, <1).
        method : str, default='iso-pdf'
            Which definition of interval to use:

            * ``'iso-pdf'``: Calculate iso probability density interval with
              the same probability density at each end. Also known as
              waterline-interval or highest average posterior density interval.
              This is only accurate if the distribution is sufficiently
              uni-modal.
            * ``'lower-limit'``/``'upper-limit'``: Lower/upper limit. One-sided
              limits for which ``level`` fraction of the (equally weighted)
              samples lie above/below the limit.
            * ``'equal-tailed'``: Equal-tailed interval with the same fraction
              of (equally weighted) samples below and above the interval
              region.

        return_covariance: bool, default=False
            Return the covariance of the sampled limits, in addition to the
            mean
        nsamples : int, default=12
            Number of CDF samples to improve `mean` and `std` estimate.

        Returns
        -------
        limit(s) : float, array, or tuple of floats or arrays
            Returns the credibility interval boundaries for each column.
            By default, returns the mean over ``nsamples`` samples, which is
            either two numbers (``method='iso-pdf'``/``'equal-tailed'``) or
            one number (``method='lower-limit'``/``'upper-limit'``). If
            ``return_covariance=True``, returns a tuple (means, covariances)
            where covariances are the covariance over the sampled limits for
            each column.
        """
        if 'lower' in method:
            limits = ['lower']
        elif 'upper' in method:
            limits = ['upper']
        else:
            limits = ['lower', 'upper']
        cis = [credibility_interval(self[col], weights=self.get_weights(),
                                    level=level, method=method,
                                    return_covariance=return_covariance,
                                    nsamples=nsamples) for col in self.columns]
        if return_covariance:
            cis, covs = zip(*cis)
            mulidx = MultiIndex.from_product([
                self.columns.get_level_values(level=0),
                limits
            ])
            ncol = len(self.columns)
            nlim = len(limits)
            covs = np.asarray(covs).reshape(nlim*ncol, nlim).T
            covs = DataFrame(covs, index=limits, columns=mulidx)
        cis = np.atleast_2d(cis) if 'limit' in method else np.asarray(cis).T
        cis = DataFrame(data=cis, index=limits, columns=self.columns)
        if return_covariance:
            return cis, covs
        else:
            return cis

    @property
    def _constructor_sliced(self):
        return WeightedSeries

    @property
    def _constructor(self):
        return WeightedDataFrame

    def groupby(
        self,
        by=None,
        axis=no_default,
        level=None,
        as_index: bool = True,
        sort: bool = True,
        group_keys: bool = True,
        observed: bool = False,
        dropna: bool = True,
    ):  # pragma: no cover  # noqa: D102
        if axis is not lib.no_default:
            axis = self._get_axis_number(axis)
            if axis == 1:
                warnings.warn(
                    "DataFrame.groupby with axis=1 is deprecated. Do "
                    "`frame.T.groupby(...)` without axis instead.",
                    FutureWarning,
                    stacklevel=find_stack_level(),
                )
            else:
                warnings.warn(
                    "The 'axis' keyword in DataFrame.groupby is deprecated "
                    "and will be removed in a future version.",
                    FutureWarning,
                    stacklevel=find_stack_level(),
                )
        else:
            axis = 0

        if level is None and by is None:
            raise TypeError("You have to supply one of 'by' and 'level'")

        return WeightedDataFrameGroupBy(
            obj=self,
            keys=by,
            level=level,
            as_index=as_index,
            sort=sort,
            group_keys=group_keys,
            observed=observed,
            dropna=dropna,
        )


for cls in [WeightedDataFrame, WeightedSeries, WeightedGroupBy,
            WeightedDataFrameGroupBy, WeightedSeriesGroupBy]:
    adjust_docstrings(cls, r'\bDataFrame\b', 'WeightedDataFrame')
    adjust_docstrings(cls, r'\bDataFrames\b', 'WeightedDataFrames')
    adjust_docstrings(cls, r'\bSeries\b', 'WeightedSeries')
    adjust_docstrings(cls, r'\bDataFrameGroupBy\b', 'WeightedDataFrameGroupBy')
    adjust_docstrings(cls, r'\bSeriesGroupBy\b', 'WeightedSeriesGroupBy')
    adjust_docstrings(cls, 'core.window.ewm', 'pandas.api.typing')
    adjust_docstrings(cls, 'core.window.expanding', 'pandas.api.typing')
    adjust_docstrings(cls, 'core.window.rolling', 'pandas.api.typing')
    adjust_docstrings(cls, 'core.window', 'pandas.api.typing')
adjust_docstrings(WeightedDataFrame, 'resample', 'pandas.DataFrame.resample')
adjust_docstrings(WeightedSeries,    'resample', 'pandas.Series.resample')



import pandas.plotting._matplotlib.boxplot
from pandas.plotting._matplotlib.boxplot import (BoxPlot as _BoxPlot,
                                                 maybe_color_bp)
from anesthetic.plotting._matplotlib.core import _WeightedMPLPlot, _get_weights
from anesthetic.utils import quantile
from pandas.core.dtypes.missing import remove_na_arraylike
import numpy as np
from pandas.io.formats.printing import pprint_thing
from anesthetic._code_utils import replace_inner_function


def _bxpstat(y, weights, whis):
    q1, med, q3 = quantile(y, [0.25, 0.5, 0.75], weights)
    bxpstat = {}
    bxpstat['med'] = med
    bxpstat['q1'] = q1
    bxpstat['q3'] = q3
    bxpstat['whislo'] = q1 - whis*(q3-q1)
    bxpstat['whishi'] = q3 + whis*(q3-q1)
    return bxpstat


class BoxPlot(_WeightedMPLPlot, _BoxPlot):
    # noqa: disable=D101
    @classmethod
    def _plot(cls, ax, y, column_num=None, return_type="axes", **kwds):
        if y.ndim == 2:
            y = [remove_na_arraylike(v) for v in y]
            # Boxplot fails with empty arrays, so need to add a NaN
            #   if any cols are empty
            # GH 8181
            y = [v if v.size > 0 else np.array([np.nan]) for v in y]
        else:
            y = remove_na_arraylike(y)

        weights = kwds.pop("weights", None)

        if weights is None:
            bp = ax.boxplot(y, **kwds)
        else:
            whis = kwds.pop("whis", 1.5)
            kwds['showfliers'] = False
            y = np.atleast_2d(y)
            bp = ax.bxp([_bxpstat(yi, weights, whis) for yi in y], **kwds)

        if return_type == "dict":
            return bp, bp
        elif return_type == "both":
            return cls.BP(ax=ax, lines=bp), bp
        else:
            return ax, bp


def boxplot(data, *args, **kwds):
    # noqa: disable=D103
    _get_weights(kwds, data)

    def create_plot_group():
        fontsize = None  # pragma: no cover
        return_type = None  # pragma: no cover
        rot = None  # pragma: no cover
        colors = None  # pragma: no cover

        def plot_group(keys, values, ax, **kwds):  # pragma: no cover
            # GH 45465: xlabel/ylabel need to be popped out before plotting
            xlabel = kwds.pop("xlabel", None)
            ylabel = kwds.pop("ylabel", None)
            if xlabel:
                ax.set_xlabel(pprint_thing(xlabel))
            if ylabel:
                ax.set_ylabel(pprint_thing(ylabel))

            weights = kwds.pop("weights", None)
            keys = [pprint_thing(x) for x in keys]
            if weights is None:
                values = [np.asarray(remove_na_arraylike(v), dtype=object)
                          for v in values]
                bp = ax.boxplot(values, **kwds)
            else:
                whis = kwds.pop("whis", 1.5)
                kwds['showfliers'] = False
                from anesthetic.plotting._matplotlib.boxplot import _bxpstat
                bp = ax.bxp([_bxpstat(v, weights, whis) for v in values],
                            **kwds)

            if fontsize is not None:
                ax.tick_params(axis="both", labelsize=fontsize)

            # GH 45465: x/y are flipped when "vert" changes
            is_vertical = (kwds.get('orientation', 'vertical') == 'vertical'
                           and kwds.get("vert", True))
            ticks = ax.get_xticks() if is_vertical else ax.get_yticks()
            if len(ticks) != len(keys):
                i, remainder = divmod(len(ticks), len(keys))
                assert remainder == 0, remainder
                keys *= i
            if is_vertical:
                ax.set_xticklabels(keys, rotation=rot)
            else:
                ax.set_yticklabels(keys, rotation=rot)
            maybe_color_bp(bp, color_tup=colors, **kwds)

            # Return axes in multiplot case, maybe revisit later # 985
            if return_type == "dict":
                return bp
            elif return_type == "both":
                return BoxPlot.BP(ax=ax, lines=bp)
            else:
                return ax

    boxplot = replace_inner_function(pandas.plotting._matplotlib.boxplot,
                                     create_plot_group.__code__.co_consts[1])
    return boxplot(data, *args, **kwds)


def boxplot_frame(data, *args, **kwds):
    # noqa: disable=D103
    _get_weights(kwds, data)
    import matplotlib.pyplot as plt

    ax = boxplot(data, *args, **kwds)
    plt.draw_if_interactive()
    return ax



import numpy as np
from numpy.testing import assert_allclose
from anesthetic.examples.utils import (
    random_ellipsoid, random_covariance, volume_n_ball, log_volume_n_ball
)
from anesthetic.examples.perfect_ns import (
    gaussian, correlated_gaussian, wedding_cake, planck_gaussian
)
from scipy.spatial import ConvexHull
from scipy.special import logsumexp


def test_random_covariance():
    np.random.seed(0)
    d = 5
    sigmas = np.random.rand(d)
    cov = random_covariance(sigmas)
    assert cov.shape == (d, d)
    e = np.linalg.eigvals(cov)
    assert_allclose(np.sort(e), np.sort(sigmas**2))


def test_random_ellipsoid():
    np.random.seed(0)

    d = 5
    mean = np.random.rand(d)
    cov = random_covariance(np.random.rand(d))

    def rvs(shape=None):
        return random_ellipsoid(mean, cov, shape)

    assert rvs(10).shape == (10, d)
    assert rvs((10,)).shape == (10, d)
    assert rvs((10, 20)).shape == (10, 20, d)

    dat = rvs(100000)

    atol = 1/np.sqrt(len(dat))
    assert_allclose(np.mean(dat, axis=0), mean, atol=atol)
    assert_allclose(np.cov(dat.T), cov/(d+2), atol=atol)

    cov = random_covariance(np.random.rand(2))
    mean = [0, 0]
    dat = rvs(1000000)
    hull = ConvexHull(dat)
    vol = hull.volume
    assert_allclose(vol, np.linalg.det(cov)**0.5*volume_n_ball(2), rtol=1e-3)


def test_perfect_ns_gaussian():
    np.random.seed(0)
    nlive = 500
    ndims = 10
    sigma = 0.1
    R = 1

    samples = gaussian(nlive, ndims, sigma, R)

    assert (samples[:-nlive].nlive == nlive).all()

    mean = samples[np.arange(ndims)].mean()
    assert_allclose(mean, 0, atol=1e-2)
    cov = samples[np.arange(ndims)].cov()
    assert_allclose(cov, np.eye(ndims)*sigma**2, atol=1e-3)

    D_KL = log_volume_n_ball(ndims) - ndims/2*np.log(2*np.pi*np.e*sigma**2)
    assert_allclose(samples.D_KL(), D_KL, atol=3*samples.D_KL(12).std())

    assert_allclose(samples.d_G(), ndims, atol=3*samples.d_G(12).std())

    logZ = ndims/2 * np.log(2*np.pi*sigma**2) - log_volume_n_ball(ndims, R)
    assert_allclose(samples.logZ(), logZ, atol=3*samples.logZ(12).std())


def test_perfect_ns_correlated_gaussian():
    np.random.seed(0)
    nlive = 500
    ndims = 5
    mean = 0.5*np.ones(ndims)
    cov = random_covariance(np.random.rand(ndims)*0.1)
    samples = correlated_gaussian(nlive, mean, cov)
    samples.gui()
    logZ = np.linalg.slogdet(2*np.pi*cov)[1]/2
    assert_allclose(samples.logZ(), logZ, atol=3*samples.logZ(12).std())
    assert (samples[:-nlive].nlive >= nlive).all()
    assert_allclose(samples[np.arange(ndims)].mean(), mean, atol=1e-2)
    assert_allclose(samples[np.arange(ndims)].cov(), cov/(ndims+2), atol=1e-2)


def test_wedding_cake():
    np.random.seed(0)
    nlive = 500
    ndims = 4
    sigma = 0.01
    alpha = 0.5
    samples = wedding_cake(nlive, ndims, sigma, alpha)

    assert samples.nlive.iloc[0] == nlive
    assert samples.nlive.iloc[-1] == 1
    assert (samples.nlive <= nlive).all()

    mean = np.sqrt(ndims/2) * (np.log(4*ndims*sigma**2)-1) / np.log(alpha)
    std = -np.sqrt(ndims/2)/np.log(alpha)
    i = np.arange(mean + std*10)
    logZ = logsumexp(-alpha**(2*i/ndims)/8/sigma**2
                     + i*np.log(alpha) + np.log(1-alpha))

    assert_allclose(samples.logZ(), logZ, atol=3*samples.logZ(12).std())


def test_planck():
    np.random.seed(0)
    nlive = 50
    cov = np.array([
        [2.12e-08, -9.03e-08, 1.76e-08, 2.96e-07, 4.97e-07, 2.38e-07],
        [-9.03e-08, 1.39e-06, -1.26e-07, -3.41e-06, -4.15e-06, -3.28e-06],
        [1.76e-08, -1.26e-07, 9.71e-08, 4.30e-07, 7.41e-07, 4.13e-07],
        [2.96e-07, -3.41e-06, 4.30e-07, 5.33e-05, 9.53e-05, 1.05e-05],
        [4.97e-07, -4.15e-06, 7.41e-07, 9.53e-05, 2.00e-04, 1.35e-05],
        [2.38e-07, -3.28e-06, 4.13e-07, 1.05e-05, 1.35e-05, 1.73e-05]])

    mean = np.array([0.02237, 0.1200, 1.04092, 0.0544, 3.044, 0.9649])

    bounds = np.array([
        [5.00e-03, 1.00e-01],
        [1.00e-03, 9.90e-01],
        [5.00e-01, 1.00e+01],
        [1.00e-02, 8.00e-01],
        [1.61e+00, 3.91e+00],
        [8.00e-01, 1.20e+00]])

    logL_mean = -1400.35
    d = 6
    logdetroottwopicov = np.linalg.slogdet(2*np.pi*cov)[1]/2
    logZ = logL_mean + logdetroottwopicov + d/2
    logV = np.log(np.diff(bounds)).sum()
    D_KL = logV - logdetroottwopicov - d/2
    planck = planck_gaussian(nlive)

    params = ['omegabh2', 'omegach2', 'theta', 'tau', 'logA', 'ns']

    assert (planck[params] > bounds.T[0]).all().all()
    assert (planck[params] < bounds.T[1]).all().all()
    assert_allclose(planck[params].mean(), mean, atol=1e-3)
    assert_allclose(planck[params].cov(), cov, atol=1e-3)
    assert_allclose(planck.logL.mean(), logL_mean,
                    atol=3*planck.logL_P(12).std())
    assert_allclose(planck.logZ(), logZ, atol=3*planck.logZ(12).std())
    assert_allclose(planck.D_KL(), D_KL, atol=3*planck.D_KL(12).std())
    assert_allclose(planck.d_G(), len(params), atol=3*planck.d_G(12).std())
    assert_allclose(planck.logL_P(), logL_mean, atol=3*planck.logL_P(12).std())


def test_logLmax():
    nlive = 1000
    mean = [0.1, 0.3, 0.5]
    # cov = covariance matrix
    cov = np.array([[.01, 0.009, 0], [0.009, .01, 0], [0, 0, 0.1]])*0.01
    bounds = [[0, 1], [0, 1], [0, 1]]
    logLmax = 10
    # d = number of parameters
    d = len(mean)
    samples = correlated_gaussian(nlive, mean, cov, bounds=bounds,
                                  logLmax=logLmax)
    abs_err = samples.logL.std()/np.sqrt(samples.neff())
    atol = abs_err*3
    assert_allclose(samples.logL.mean(), logLmax-d/2, atol=atol)



"""Anesthetic testing utilities."""
import pandas.testing
import numpy.testing


def assert_frame_equal(left, right, *args, **kwargs):
    """Assert frames are equal, including metadata."""
    check_metadata = kwargs.pop('check_metadata', True)
    pandas.testing.assert_frame_equal(left, right, *args, **kwargs)
    numpy.testing.assert_array_equal(left._metadata, right._metadata)
    if check_metadata:
        for key in left._metadata:
            assert getattr(left, key) == getattr(right, key)



from anesthetic.weighted_pandas import (WeightedDataFrame, WeightedSeries,
                                        read_csv)
from pandas import DataFrame, MultiIndex
import pandas.testing
from anesthetic.utils import neff
import pytest
from pytest import approx
import numpy as np
from numpy.testing import assert_array_equal, assert_allclose
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.colors import to_rgba
from pandas.plotting import scatter_matrix, bootstrap_plot
from pandas.plotting._matplotlib.misc import (
    scatter_matrix as orig_scatter_matrix
)
from packaging import version


@pytest.fixture(autouse=True)
def close_figures_on_teardown():
    yield
    plt.close("all")


@pytest.fixture
def series():
    np.random.seed(0)
    N = 100000
    data = np.random.rand(N)

    series = WeightedSeries(data)
    assert_array_equal(series.get_weights(), 1)
    assert not series.isweighted()
    assert_array_equal(series, data)

    series = WeightedSeries(data, weights=None)
    assert_array_equal(series.get_weights(), 1)
    assert not series.isweighted()
    assert_array_equal(series, data)

    weights = np.random.rand(N)
    series = WeightedSeries(data, weights=weights)
    assert series.isweighted()
    assert_array_equal(series, data)

    assert series.get_weights().shape == (N,)
    assert series.shape == (N,)
    assert isinstance(series.get_weights(), np.ndarray)
    assert_array_equal(series, data)
    assert_array_equal(series.get_weights(), weights)
    assert isinstance(series.to_frame(), WeightedDataFrame)
    assert_array_equal(series.to_frame().get_weights(), weights)
    assert_array_equal(series.index.get_level_values('weights'), weights)

    return series


@pytest.fixture
def frame():
    np.random.seed(0)
    N = 100000
    cols = ['A', 'B', 'C', 'D', 'E', 'F']
    m = len(cols)
    data = np.random.rand(N, m)

    frame = WeightedDataFrame(data, columns=cols)
    assert_array_equal(frame.get_weights(), 1)
    assert not frame.isweighted(0) and not frame.isweighted(1)
    assert_array_equal(frame, data)

    frame = WeightedDataFrame(data, weights=None, columns=cols)
    assert_array_equal(frame.get_weights(), 1)
    assert not frame.isweighted(0) and not frame.isweighted(1)
    assert_array_equal(frame, data)

    weights = np.random.rand(N)
    frame = WeightedDataFrame(data, weights=weights, columns=cols)
    assert frame.isweighted(0) and not frame.isweighted(1)
    assert frame.get_weights().shape == (N,)
    assert frame.shape == (N, m)
    assert isinstance(frame.get_weights(), np.ndarray)
    assert_array_equal(frame, data)
    assert_array_equal(frame.get_weights(), weights)
    assert_array_equal(frame.columns, cols)
    assert_array_equal(frame.index.get_level_values('weights'), weights)
    return frame


@pytest.fixture(params=["frame", "series"])
def samples(request, frame, series):
    return frame if request.param == "frame" else series


def test_WeightedDataFrame_key(frame):
    for key1 in frame.columns:
        assert_array_equal(frame.get_weights(), frame[key1].get_weights())
        for key2 in frame.columns:
            assert_array_equal(frame[key1].get_weights(),
                               frame[key2].get_weights())


def test_WeightedDataFrame_slice(frame):
    assert isinstance(frame['A'], WeightedSeries)
    assert isinstance(frame.iloc[0], WeightedSeries)
    assert not frame.iloc[0].isweighted()
    assert isinstance(frame[:10], WeightedDataFrame)
    assert frame[:10].isweighted()
    assert frame[:10].shape == (10, 6)
    assert frame[:10].get_weights().shape == (10,)
    assert frame[:10]._rand(0).shape == (10,)
    assert frame[:10]._rand(1).shape == (6,)


def test_WeightedDataFrame_mean(frame):
    mean = frame.mean()
    assert isinstance(mean, WeightedSeries)
    assert not mean.isweighted()
    assert_array_equal(mean.index, frame.columns)
    assert_allclose(mean, 0.5, atol=1e-2)

    mean = frame.mean(axis=1)
    assert isinstance(mean, WeightedSeries)
    assert_array_equal(mean, frame.T.mean())
    assert isinstance(frame.T.mean(), WeightedSeries)
    assert mean.isweighted()
    assert frame.T.mean().isweighted()
    assert_array_equal(mean.index, frame.index)
    assert_allclose(mean.mean(), 0.5, atol=1e-2)


def test_WeightedDataFrame_std(frame):
    std = frame.std()
    assert isinstance(std, WeightedSeries)
    assert not std.isweighted()
    assert_array_equal(std.index, frame.columns)
    assert_allclose(std, (1./12)**0.5, atol=1e-2)
    assert np.all(frame.std(ddof=0) != frame.std(ddof=1))

    std = frame.std(axis=1)
    assert isinstance(std, WeightedSeries)
    assert_array_equal(std, frame.T.std())
    assert isinstance(frame.T.std(), WeightedSeries)
    assert std.isweighted()
    assert frame.T.std().isweighted()
    assert_array_equal(std.index, frame.index)
    assert_allclose(std.mean(), (1./12)**0.5, atol=1e-1)


def test_WeightedDataFrame_cov(frame):
    cov = frame.cov()
    assert isinstance(cov, WeightedDataFrame)
    assert not cov.isweighted(0) and not cov.isweighted(1)
    assert_array_equal(cov.index, frame.columns)
    assert_array_equal(cov.columns, frame.columns)
    assert_allclose(cov, (1./12)*np.identity(6), atol=1e-2)

    cov = frame[:5].T.cov()
    assert isinstance(cov, WeightedDataFrame)
    assert cov.isweighted(0) and cov.isweighted(1)
    assert_array_equal(cov.index, frame[:5].index)
    assert_array_equal(cov.columns, frame[:5].index)


def test_WeightedDataFrame_corr(frame):
    corr = frame.corr()
    assert isinstance(corr, WeightedDataFrame)
    assert not corr.isweighted(0) and not corr.isweighted(1)
    assert_array_equal(corr.index, frame.columns)
    assert_array_equal(corr.columns, frame.columns)
    assert_allclose(corr, np.identity(6), atol=1.1e-2)

    corr = frame[:5].T.corr()
    assert isinstance(corr, WeightedDataFrame)
    assert corr.isweighted(0) and corr.isweighted(1)
    assert_array_equal(corr.index, frame[:5].index)
    assert_array_equal(corr.columns, frame[:5].index)


def test_WeightedDataFrame_corrwith(frame):
    correl = frame.corrwith(frame.A)
    assert isinstance(correl, WeightedSeries)
    assert not correl.isweighted()
    assert_array_equal(correl.index, frame.columns)
    assert_allclose(correl, frame.corr()['A'], rtol=1e-10, atol=1e-10)

    correl = frame.corrwith(frame[['A', 'B']])
    assert isinstance(correl, WeightedSeries)
    assert not correl.isweighted()
    assert_array_equal(correl.index, frame.columns)
    assert_allclose(correl['A'], 1, rtol=1e-12, atol=1e-12)
    assert_allclose(correl['B'], 1, rtol=1e-12, atol=1e-12)
    assert np.isnan(correl['C'])

    unweighted = DataFrame(frame).droplevel('weights')

    if version.parse(pandas.__version__) >= version.parse('3.0.0'):
        correl = frame.corrwith(unweighted.A)
        assert_allclose(correl['A'], 1, atol=1e-2)
        assert_allclose(correl['B'], 0, atol=1e-2)

        correl = frame.corrwith(unweighted[['A', 'B']])
        assert_allclose(correl['A'], 1, atol=1e-2)
        assert_allclose(correl['B'], 1, atol=1e-2)
        assert np.isnan(correl['C'])

        correl = unweighted.corrwith(frame[['A', 'B']])
        assert_allclose(correl['A'], 1, atol=1e-2)
        assert_allclose(correl['B'], 1, atol=1e-2)
        assert np.isnan(correl['C'])
    else:
        with pytest.raises(ValueError):
            frame.corrwith(unweighted.A)

        with pytest.raises(ValueError):
            frame.corrwith(unweighted[['A', 'B']])

        with pytest.raises(ValueError):
            unweighted.corrwith(frame[['A', 'B']])

    correl_1 = unweighted[:5].corrwith(unweighted[:4], axis=1)
    correl_2 = frame[:5].corrwith(frame[:4], axis=1)
    # Allow for difference due to ddof=0 vs ddof=1 approach
    assert_allclose(correl_1.values[:-1], correl_2.values[:-1], rtol=1e-12)
    assert np.isnan(correl_1.values[-1]) and np.isnan(correl_2.values[-1])
    assert correl_2.isweighted()

    correl_3 = frame[:5].T.corrwith(frame[:4].T)
    assert_array_equal(correl_2, correl_3)
    assert_array_equal(correl_2.index, correl_3.index)

    correl_4 = frame.T.corrwith(frame.T, axis=1)
    correl_5 = unweighted.T.corrwith(unweighted.T, axis=1)
    assert_allclose(correl_4, correl_5, rtol=1e-12)

    frame.set_weights(None, inplace=True)
    assert_allclose(frame.corrwith(frame), unweighted.corrwith(unweighted),
                    rtol=1e-12)


def test_WeightedDataFrame_median(frame):
    median = frame.median()
    assert isinstance(median, WeightedSeries)
    assert not median.isweighted()
    assert_array_equal(median.index, frame.columns)
    assert_allclose(median, 0.5, atol=1e-2)

    median = frame.median(axis=1)
    assert isinstance(median, WeightedSeries)
    assert_array_equal(median, frame.T.median())
    assert isinstance(frame.T.median(), WeightedSeries)
    assert median.isweighted()
    assert frame.T.median().isweighted()
    assert_array_equal(median.index, frame.index)
    assert_allclose(median.mean(), 0.5, atol=1e-2)


def test_WeightedDataFrame_sem(frame):
    sem = frame.sem()
    assert isinstance(sem, WeightedSeries)
    assert not sem.isweighted()
    assert_array_equal(sem.index, frame.columns)
    assert_allclose(sem, (1./12)**0.5/np.sqrt(frame.neff()), atol=1e-2)
    assert np.all(frame.sem(ddof=0) != frame.sem(ddof=1))

    sem = frame.sem(axis=1)
    assert isinstance(sem, WeightedSeries)
    assert_array_equal(sem, frame.T.sem())
    assert isinstance(frame.T.sem(), WeightedSeries)
    assert sem.isweighted()
    assert frame.T.sem().isweighted()
    assert_array_equal(sem.index, frame.index)


def test_WeightedDataFrame_kurtosis(frame):
    kurtosis = frame.kurtosis()
    assert isinstance(kurtosis, WeightedSeries)
    assert not kurtosis.isweighted()
    assert_array_equal(kurtosis.index, frame.columns)
    assert_allclose(kurtosis, -6/5, atol=1e-2)
    assert_array_equal(frame.kurtosis(), frame.kurt())

    kurtosis = frame.kurtosis(axis=1)
    assert isinstance(kurtosis, WeightedSeries)
    assert_array_equal(kurtosis, frame.T.kurtosis())
    assert isinstance(frame.T.kurtosis(), WeightedSeries)
    assert kurtosis.isweighted()
    assert frame.T.kurtosis().isweighted()
    assert_array_equal(kurtosis.index, frame.index)
    assert_array_equal(frame.kurtosis(axis=1), frame.kurt(axis=1))


def test_WeightedDataFrame_skew(frame):
    skew = frame.skew()
    assert isinstance(skew, WeightedSeries)
    assert not skew.isweighted()
    assert_array_equal(skew.index, frame.columns)
    assert_allclose(skew, 0., atol=2e-2)

    skew = frame.skew(axis=1)
    assert isinstance(skew, WeightedSeries)
    assert_array_equal(skew, frame.T.skew())
    assert isinstance(frame.T.skew(), WeightedSeries)
    assert skew.isweighted()
    assert frame.T.skew().isweighted()
    assert_array_equal(skew.index, frame.index)


def test_WeightedDataFrame_quantile(frame):
    quantile = frame.quantile()
    assert isinstance(quantile, WeightedSeries)
    assert not quantile.isweighted()
    assert_array_equal(quantile.index, frame.columns)
    assert_allclose(quantile, 0.5, atol=1e-2)

    quantile = frame.quantile(axis=1)
    assert isinstance(quantile, WeightedSeries)
    assert_array_equal(quantile, frame.T.quantile())
    assert quantile.isweighted()
    assert frame.T.quantile().isweighted()
    assert_array_equal(quantile.index, frame.index)
    assert_allclose(quantile.mean(), 0.5, atol=1e-2)

    qs = np.linspace(0, 1, 10)
    for q in qs:
        quantile = frame.quantile(q)
        assert isinstance(quantile, WeightedSeries)
        assert not quantile.isweighted()
        assert_array_equal(quantile.index, frame.columns)
        assert_allclose(quantile, q, atol=1e-2)

        quantile = frame.quantile(q, axis=1)
        assert isinstance(quantile, WeightedSeries)

        quantile = frame.quantile(q, axis=1)
        assert isinstance(quantile, WeightedSeries)
        assert_array_equal(quantile, frame.T.quantile(q))
        assert quantile.isweighted()
        assert frame.T.quantile(q).isweighted()
        assert_array_equal(quantile.index, frame.index)

    quantile = frame.quantile(qs)
    assert_allclose(quantile, np.transpose([qs]*6), atol=1e-2)
    assert_array_equal(quantile.index, qs)
    assert_array_equal(quantile.columns, frame.columns)
    quantile = frame.quantile(qs, axis=1)
    assert isinstance(quantile, WeightedDataFrame)
    assert_array_equal(quantile.index, qs)
    assert_array_equal(quantile.columns, frame.index)

    with pytest.raises(NotImplementedError):
        frame.quantile(numeric_only=False)
    with pytest.raises(NotImplementedError):
        frame.quantile(method='single')


def test_WeightedDataFrame_sample(frame):
    sample = frame.sample()
    assert isinstance(sample, WeightedDataFrame)
    assert_array_equal(sample.columns, frame.columns)
    assert sample.index.isin(frame.index)
    assert sample.isin(frame).all().all()

    samples = frame.sample(5)
    assert isinstance(samples, WeightedDataFrame)
    assert_array_equal(sample.columns, frame.columns)
    assert samples.index.isin(frame.index).all()
    assert samples.isin(frame).all().all()
    assert len(samples) == 5

    sample = frame.sample(axis=1)
    assert isinstance(sample, WeightedDataFrame)
    assert_array_equal(sample.index, frame.index)
    assert sample.columns.isin(frame.columns).all()
    assert sample.isin(frame).all().all()

    samples = frame.sample(5, axis=1)
    assert isinstance(samples, WeightedDataFrame)
    assert_array_equal(samples.index, frame.index)
    assert samples.columns.isin(frame.columns).all()
    assert samples.isin(frame).all().all()
    assert len(samples.columns) == 5

    frame.T.sample()
    frame.T.sample(5)
    frame.T.sample(axis=1)
    frame.T.sample(5, axis=1)


def test_WeightedDataFrame_neff(frame):
    N_eff = frame.neff()
    assert isinstance(N_eff, float)
    assert N_eff < len(frame)
    assert N_eff > len(frame) * np.exp(-0.25)

    N_eff = frame.neff(1)
    assert isinstance(N_eff, int)
    assert N_eff == len(frame.T)

    # beta kwarg
    for beta in [0.5, 1, 2, np.inf, '0.5', 'equal', 'entropy', 'kish']:
        assert frame.neff(beta=beta) == neff(frame.get_weights(), beta=beta)


@pytest.mark.parametrize('weighted', [False, True])
def test_weighted_compress(samples, weighted):
    no_comp = samples.compress(ncompress=False, weighted=weighted)
    equally = samples.compress(ncompress='equal', weighted=weighted)
    entropy = samples.compress(ncompress='entropy', weighted=weighted)
    small_i = samples.compress(ncompress=1000, weighted=weighted)  # small int
    large_i = samples.compress(ncompress=80000, weighted=weighted)  # large int
    small_f = samples.compress(ncompress=1e3, weighted=weighted)  # small float
    large_f = samples.compress(ncompress=8e4, weighted=weighted)  # large float

    # no compression if ncompress is False
    assert no_comp.isweighted()
    assert_array_equal(no_comp, samples)
    assert_array_equal(no_comp.get_weights(), samples.get_weights())

    if weighted:
        # weighted after compression
        assert equally.isweighted()
        assert entropy.isweighted()
        assert small_i.isweighted()
        assert large_i.isweighted()
        assert small_f.isweighted()
        assert large_f.isweighted()
        # uniform weights for small samples
        assert_allclose(equally.get_weights(), 1)
        assert_allclose(small_i.get_weights(), 1)
        assert_allclose(small_f.get_weights(), 1)
        # length tests
        assert entropy.get_weights().sum() == approx(samples.neff(), abs=200)
        assert small_i.get_weights().sum() == 1000  # exact for integers
        assert large_i.get_weights().sum() == 80000  # exact for integers
        assert small_f.get_weights().sum() == approx(1e3, abs=100)
        assert large_f.get_weights().sum() == approx(8e4, abs=200)
        assert len(np.unique(entropy.index)) == len(entropy) < len(samples)
        assert len(np.unique(large_i.index)) == len(large_i) < len(samples)
        assert len(np.unique(large_f.index)) == len(large_f) < len(samples)
    else:
        # unweighted after compression
        assert not equally.isweighted()
        assert not entropy.isweighted()
        assert not small_i.isweighted()
        assert not large_i.isweighted()
        assert not small_f.isweighted()
        assert not large_f.isweighted()
        # length tests
        assert len(entropy) == approx(samples.neff(), abs=200)
        assert len(small_i) == 1000  # exact for integers
        assert len(large_i) == 80000  # exact for integers
        assert len(small_f) == approx(1e3, abs=100)  # approximate for float
        assert len(large_f) == approx(8e4, abs=200)  # approximate for float
        assert len(np.unique(entropy.index)) < len(entropy) < len(samples)
        assert len(np.unique(large_i.index)) < len(large_i) < len(samples)
        assert len(np.unique(large_f.index)) < len(large_f) < len(samples)
    assert len(np.unique(equally.index)) == len(equally) < len(entropy)
    assert len(np.unique(small_i.index)) == len(small_i) < len(large_i)
    assert len(np.unique(small_f.index)) == len(small_f) < len(large_f)

    # ensure reproducibility
    assert_array_equal(samples.compress(weighted=weighted),
                       samples.compress(weighted=weighted))
    assert_array_equal(samples.compress(1000, weighted=weighted),
                       samples.compress(1000, weighted=weighted))
    assert_array_equal(samples.compress('equal', weighted=weighted),
                       samples.compress('equal', weighted=weighted))


def test_WeightedDataFrame_compress(frame):
    assert_array_equal(frame.T.compress().T, frame)
    assert_array_equal(frame.T.compress(axis=1).T, frame.compress())


@pytest.mark.parametrize('ncompress, expected', [(100, 100),
                                                 (1000, 1000),
                                                 (True, 1000),
                                                 (False, 1000),
                                                 ('equal', 1000)])
def test_unweighted_compress(samples, ncompress, expected):
    """Test compression works for unweighted `WeightedDataFrames`."""
    # Create unweighted data
    np.random.seed(42)
    samples = samples[:1000].drop_weights()
    assert not samples.isweighted()

    # Test compression
    compressed = samples.compress(ncompress)
    assert not compressed.isweighted()
    assert len(compressed) == expected
    if isinstance(samples, DataFrame):
        assert list(compressed.columns) == list(samples.columns)
    else:
        assert compressed.name == samples.name


def test_WeightedDataFrame_nan(frame):
    frame.loc[0, 'A'] = np.nan
    assert ~frame.mean().isna().any()
    assert ~frame.mean(axis=1).isna().any()
    assert_array_equal(frame.mean(skipna=False).isna(), [True] + [False]*5)
    assert_array_equal(frame.mean(axis=1, skipna=False).isna()[0:6],
                       [True, False, False, False, False, False])

    assert ~frame.std().isna().any()
    assert ~frame.std(axis=1).isna().any()
    assert_array_equal(frame.std(skipna=False).isna(), [True] + [False]*5)
    assert_array_equal(frame.std(axis=1, skipna=False).isna()[0:6],
                       [True, False, False, False, False, False])

    frame.loc[2, 'B'] = np.nan
    assert ~frame.mean().isna().any()
    assert_array_equal(frame.mean(skipna=False).isna(),
                       [True, True] + [False]*4)
    assert_array_equal(frame.mean(axis=1, skipna=False).isna()[0:6],
                       [True, False, True, False, False, False])

    assert ~frame.std().isna().any()
    assert_array_equal(frame.std(skipna=False).isna(),
                       [True, True] + [False]*4)
    assert_array_equal(frame.std(axis=1, skipna=False).isna()[0:6],
                       [True, False, True, False, False, False])

    frame.loc[4, 'C'] = np.nan
    frame.loc[5, 'D'] = np.nan
    frame.loc[6, 'E'] = np.nan
    frame.loc[7, 'F'] = np.nan
    assert ~frame.mean().isna().any()
    assert frame.mean(skipna=False).isna().all()
    assert_array_equal(frame.mean(axis=1, skipna=False).isna()[0:6],
                       [True, False, True, False, True, True])

    assert ~frame.std().isna().any()
    assert frame.std(skipna=False).isna().all()
    assert_array_equal(frame.std(axis=1, skipna=False).isna()[0:6],
                       [True, False, True, False, True, True])

    assert_allclose(frame.mean(), 0.5, atol=1e-2)
    assert_allclose(frame.std(), (1./12)**0.5, atol=1e-2)
    assert_allclose(frame.cov(), (1./12)*np.identity(6), atol=1e-2)

    assert isinstance(frame.mean(), WeightedSeries)
    assert not frame.mean().isweighted()
    assert isinstance(frame.mean(axis=1), WeightedSeries)
    assert frame.mean(axis=1).isweighted()

    assert frame[:0].mean().isna().all()
    assert frame[:0].std().isna().all()
    assert frame[:0].median().isna().all()
    assert frame[:0].var().isna().all()
    assert frame[:0].cov().isna().all().all()
    assert frame[:0].corr().isna().all().all()
    assert frame[:0].kurt().isna().all()
    assert frame[:0].skew().isna().all()
    assert frame[:0].sem().isna().all()
    assert frame[:0].quantile().isna().all()


def test_WeightedSeries_mean(series):
    series[0] = np.nan
    series.var(skipna=False)
    mean = series.mean()
    assert isinstance(mean, float)
    assert_allclose(mean, 0.5, atol=1e-2)


def test_WeightedSeries_std(series):
    std = series.std()
    assert isinstance(std, float)
    assert_allclose(std, (1./12)**0.5, atol=1e-2)

    series[0] = np.nan
    assert ~np.isnan(series.std())
    assert np.isnan(series.std(skipna=False))


def test_WeightedSeries_cov(frame):
    assert_allclose(frame.A.cov(frame.A), 1./12, atol=1e-2)
    assert_allclose(frame.A.cov(frame.B), 0, atol=1e-2)

    frame.loc[0, 'A'] = np.nan
    assert_allclose(frame.A.cov(frame.A), 1./12, atol=1e-2)
    assert_allclose(frame.A.cov(frame.B), 0, atol=1e-2)


def test_WeightedSeries_corr(frame):
    assert_allclose(frame.A.corr(frame.A), 1., atol=1e-2)
    assert_allclose(frame.A.corr(frame.B), 0, atol=1e-2)
    D = frame.A + frame.B
    assert_allclose(frame.A.corr(D), 1/np.sqrt(2), atol=1e-2)

    unweighted = DataFrame(frame).droplevel('weights')

    if version.parse(pandas.__version__) >= version.parse('3.0.0'):
        assert_allclose(frame.A.corr(unweighted.A), 1, atol=1e-2)
        assert_allclose(frame.A.corr(unweighted.B), 0, atol=1e-2)
        assert_allclose(unweighted.A.corr(frame.B), 0, atol=1e-2)
    else:
        with pytest.raises(ValueError):
            frame.A.corr(unweighted.B)

        with pytest.raises(ValueError):
            unweighted.A.corr(frame.B)


def test_WeightedSeries_median(series):
    median = series.median()
    assert isinstance(median, float)
    assert_allclose(median, 0.5, atol=1e-2)


def test_WeightedSeries_sem(series):
    sem = series.sem()
    assert isinstance(sem, float)
    assert_allclose(sem, (1./12)**0.5/np.sqrt(series.neff()), atol=1e-2)


def test_WeightedSeries_kurtosis(series):
    kurtosis = series.kurtosis()
    assert isinstance(kurtosis, float)
    assert_allclose(kurtosis, -6/5, atol=1e-2)
    assert series.kurtosis() == series.kurt()

    series[0] = np.nan
    assert ~np.isnan(series.kurtosis())
    assert np.isnan(series.kurtosis(skipna=False))


def test_WeightedSeries_skew(series):
    skew = series.skew()
    assert isinstance(skew, float)
    assert_allclose(skew, 0., atol=1e-2)

    series[0] = np.nan
    assert ~np.isnan(series.skew())
    assert np.isnan(series.skew(skipna=False))


def test_WeightedSeries_quantile(series):

    quantile = series.quantile()
    assert isinstance(quantile, float)
    assert_allclose(quantile, 0.5, atol=1e-2)

    qs = np.linspace(0, 1, 10)
    for q in qs:
        quantile = series.quantile(q)
        assert isinstance(quantile, float)
        assert_allclose(quantile, q, atol=1e-2)

    assert_allclose(series.quantile(qs), qs, atol=1e-2)


def test_WeightedSeries_sample(series):
    sample = series.sample()
    assert isinstance(sample, WeightedSeries)
    samples = series.sample(5)
    assert isinstance(samples, WeightedSeries)
    assert len(samples) == 5


def test_WeightedSeries_neff(series):
    neff = series.neff()
    assert isinstance(neff, float)
    assert neff < len(series)
    assert neff > len(series) * np.exp(-0.25)


def test_WeightedSeries_nan(series):

    series[0] = np.nan

    assert ~np.isnan(series.mean())
    assert np.isnan(series.mean(skipna=False))
    assert ~np.isnan(series.std())
    assert np.isnan(series.std(skipna=False))
    assert ~np.isnan(series.var())
    assert np.isnan(series.var(skipna=False))

    assert_allclose(series.mean(), 0.5, atol=1e-2)
    assert_allclose(series.var(), 1./12, atol=1e-2)
    assert_allclose(series.std(), (1./12)**0.5, atol=1e-2)

    assert np.isnan(series[:0].mean())
    assert np.isnan(series[:0].std())
    assert np.isnan(series[:0].median())
    assert np.isnan(series[:0].var())
    assert np.isnan(series[:0].cov(series))
    assert np.isnan(series[:0].corr(series))
    assert np.isnan(series[:0].kurt())
    assert np.isnan(series[:0].skew())
    assert np.isnan(series[:0].sem())
    assert np.isnan(series[:0].quantile())


# =============================================================================
# Numerical Correctness Tests for NaN Handling
# =============================================================================

@pytest.mark.parametrize("data, weights, expected", [
    # Case 1: The original bug from Lukas's report
    ([1.0, 2.0, np.nan, 4.0], [0.1, 0.2, 0.3, 0.4], 3.0),
    # Case 2: NaN at beginning
    ([np.nan, 2.0, 3.0, 4.0], [0.1, 0.2, 0.3, 0.4], 29/9),
    # Case 3: NaN at end
    ([1.0, 2.0, 3.0, np.nan], [0.1, 0.2, 0.3, 0.4], 7/3),
    # Case 4: Multiple NaNs
    ([1.0, np.nan, 3.0, np.nan, 5.0], [0.1, 0.2, 0.3, 0.1, 0.3], 25/7),
    # Case 5: NaN with large weight (tests that weight is excluded)
    ([10.0, np.nan, 30.0], [1.0, 5.0, 2.0], 70/3),
    # Case 6: NaN with zero weight (should be same as excluding it)
    ([1.0, 2.0, np.nan], [0.5, 0.5, 0.0], 1.5),
])
def test_weighted_series_mean_nan_correctness(data, weights, expected):
    """Test that mean(skipna=True) returns numerically correct results."""
    result = WeightedSeries(data, weights=weights).mean(skipna=True)
    assert_allclose(result, expected, rtol=1e-10)


@pytest.mark.parametrize("data, weights", [
    ([1.0, 2.0, np.nan, 4.0], [0.1, 0.2, 0.3, 0.4]),
    ([np.nan, 2.0, 3.0], [0.3, 0.3, 0.4]),
    ([1.0, np.nan], [0.6, 0.4]),
])
def test_weighted_series_mean_skipna_false_returns_nan(data, weights):
    """Test that mean(skipna=False) returns NaN when NaN present."""
    result = WeightedSeries(data, weights=weights).mean(skipna=False)
    assert np.isnan(result)


@pytest.mark.parametrize("data, weights, ddof, expected", [
    ([1.0, 2.0, np.nan, 4.0], [1, 1, 1, 3], 0, 1.6),
    ([1.0, 2.0, np.nan, 4.0], [1, 1, 1, 3], 1, 2.0),
    ([np.nan, 1.0, 2.0, 4.0], [1, 1, 1, 3], 0, 1.6),
    ([np.nan, 1.0, 2.0, 4.0], [1, 1, 1, 3], 1, 2.0),
    ([1.0, np.nan, 3.0, np.nan], [1, 1, 1, 1], 0, 1.0),
    ([1.0, np.nan, 3.0, np.nan], [1, 1, 1, 1], 1, 2.0),
    ([1.0, np.nan, 3.0, np.nan], [0.25, 0.25, 0.25, 0.25], 0, 1.0),
    ([1.0, np.nan, 3.0, np.nan], [0.25, 0.25, 0.25, 0.25], 1, 2.0),
])
def test_weighted_series_var_nan_correctness(data, weights, ddof, expected):
    """Test that var(skipna=True) returns numerically correct results."""
    result = WeightedSeries(data, weights=weights).var(skipna=True, ddof=ddof)
    assert_allclose(result, expected, rtol=1e-10)


@pytest.mark.parametrize("data, weights, ddof, expected", [
    ([1.0, 2.0, np.nan, 4.0], [1, 1, 1, 3], 0, np.sqrt(1.6)),
    ([1.0, 2.0, np.nan, 4.0], [1, 1, 1, 3], 1, np.sqrt(2.0)),
    ([np.nan, 1.0, 2.0, 4.0], [1, 1, 1, 3], 0, np.sqrt(1.6)),
    ([np.nan, 1.0, 2.0, 4.0], [1, 1, 1, 3], 1, np.sqrt(2.0)),
    ([1.0, np.nan, 3.0, np.nan], [1, 1, 1, 1], 0, 1.0),
    ([1.0, np.nan, 3.0, np.nan], [1, 1, 1, 1], 1, np.sqrt(2.0)),
    ([1.0, np.nan, 3.0, np.nan], [0.25, 0.25, 0.25, 0.25], 0, 1.0),
    ([1.0, np.nan, 3.0, np.nan], [0.25, 0.25, 0.25, 0.25], 1, np.sqrt(2.0)),
])
def test_weighted_series_std_nan_correctness(data, weights, ddof, expected):
    """Test that std(skipna=True) returns numerically correct results."""
    result = WeightedSeries(data, weights=weights).std(skipna=True, ddof=ddof)
    assert_allclose(result, expected, rtol=1e-10)


def test_weighted_series_all_nan_returns_nan():
    """Test that all statistical methods return NaN for all-NaN data."""
    data = [np.nan, np.nan, np.nan, np.nan, np.nan]
    weights = [0.3, 0.3, 0.4, 0.5, 0.1]
    series = WeightedSeries(data, weights=weights)

    assert np.isnan(series.mean(skipna=True))
    assert np.isnan(series.var(skipna=True, ddof=0))
    assert np.isnan(series.var(skipna=True, ddof=1))
    assert np.isnan(series.std(skipna=True))
    assert np.isnan(series.sem(skipna=True))
    assert np.isnan(series.skew(skipna=True))
    assert np.isnan(series.kurt(skipna=True))
    assert np.isnan(series.cov(series, ddof=0))
    assert np.isnan(series.cov(series, ddof=1))
    assert np.isnan(series.corr(series))


def test_weighted_series_equivalence_to_manual_filtering():
    """Test that skipna=True equals manual filtering of data and weights."""
    data = [1.0, np.nan, 3.0, np.nan, 5.0, 6.0]
    weights = [0.1, 0.2, 0.1, 0.3, 0.2, 0.1]

    s = WeightedSeries(data, weights=weights)

    # Manual filtering approach
    mask = ~np.isnan(data)
    filtered_data = np.array(data)[mask]
    filtered_weights = np.array(weights)[mask]
    fs = WeightedSeries(filtered_data, weights=filtered_weights)

    # Results should be identical
    assert_allclose(s.mean(skipna=True), fs.mean(), rtol=1e-12)
    assert_allclose(s.var(skipna=True, ddof=0), fs.var(ddof=0), rtol=1e-12)
    assert_allclose(s.var(skipna=True, ddof=1), fs.var(ddof=1), rtol=1e-12)
    assert_allclose(s.std(skipna=True), fs.std(), rtol=1e-12)
    assert_allclose(s.sem(skipna=True), fs.sem(), rtol=1e-12)
    assert_allclose(s.skew(skipna=True), fs.skew(), rtol=1e-12)
    assert_allclose(s.kurt(skipna=True), fs.kurt(), rtol=1e-12)
    assert_allclose(s.cov(s, ddof=0), fs.cov(fs, ddof=0), rtol=1e-12)
    assert_allclose(s.cov(s, ddof=1), fs.cov(fs, ddof=1), rtol=1e-12)
    assert_allclose(s.corr(s), fs.corr(fs), rtol=1e-12)


def test_weighted_dataframe_nan_consistency():
    """Test that DataFrame.mean()[col] equals Series.mean() for same data."""
    data = [[1.0, np.nan, 3.0], [4.0, 5.0, np.nan]]
    weights = [0.4, 0.6]

    df = WeightedDataFrame(data, columns=['A', 'B', 'C'], weights=weights)
    df_result = df.mean(skipna=True)

    for col in ['A', 'B', 'C']:
        s = df[col]
        assert isinstance(s, WeightedSeries)
        series_result = s.mean(skipna=True)
        assert_allclose(df_result[col], series_result, rtol=1e-10)


def test_weighted_dataframe_mean_ground_truth():
    """Test DataFrame mean with ground truth values."""
    data = [[1.0, np.nan, 3.0],
            [4.0, 5.0, np.nan]]
    weights = [0.4, 0.6]
    expected_means = {'A': 2.8, 'B': 5.0, 'C': 3.0}

    df = WeightedDataFrame(data, columns=['A', 'B', 'C'], weights=weights)
    result = df.mean(skipna=True)

    for col, expected in expected_means.items():
        assert_allclose(result[col], expected, rtol=1e-10)


@pytest.fixture
def mcmc_df():
    np.random.seed(0)
    m, s = 0.5, 0.1

    def logL(x):
        return x, -((x-m)**2).sum(axis=-1)/2/s**2

    np.random.seed(0)
    x0, logL0 = logL(np.random.normal(m, s, 4))
    dat = []
    for _ in range(3000):
        dat.append(x0)
        x1, logL1 = logL(np.random.normal(x0, s/2))
        if np.log(np.random.rand()) < logL1 - logL0:
            x0, logL0 = x1, logL1

    return DataFrame(dat, columns=["x", "y", "z", "w"])


@pytest.fixture
def mcmc_wdf(mcmc_df):
    weights = mcmc_df.groupby(mcmc_df.columns.tolist(), sort=False).size()
    return WeightedDataFrame(mcmc_df.drop_duplicates(), weights=weights.values)


def test_mcmc_stats(mcmc_df, mcmc_wdf):
    df = mcmc_df
    wdf = mcmc_wdf
    assert_allclose(wdf.mean(), df.mean(), rtol=1e-12)
    assert_allclose(wdf.var(ddof=0), df.var(ddof=0), rtol=1e-12)
    assert_allclose(wdf.var(ddof=1), df.var(ddof=1), rtol=1e-12)
    assert_allclose(wdf.std(ddof=0), df.std(ddof=0), rtol=1e-12)
    assert_allclose(wdf.std(ddof=1), df.std(ddof=1), rtol=1e-12)
    assert_allclose(wdf.sem(ddof=0), df.sem(ddof=0), rtol=1e-12)
    assert_allclose(wdf.sem(ddof=1), df.sem(ddof=1), rtol=1e-12)
    assert_allclose(wdf.skew(), df.skew(), rtol=1e-12)
    assert_allclose(wdf.kurt(), df.kurt(), rtol=1e-12)
    assert_allclose(wdf.cov(ddof=0), df.cov(ddof=0), rtol=1e-12)
    assert_allclose(wdf.cov(ddof=1), df.cov(ddof=1), rtol=1e-12)
    assert_allclose(np.diag(wdf.cov(ddof=0)), wdf.var(ddof=0), rtol=1e-12)
    assert_allclose(np.diag(wdf.cov(ddof=1)), wdf.var(ddof=1), rtol=1e-12)
    assert_allclose(wdf.corr(), df.corr(), rtol=1e-12, atol=1e-12)
    assert_allclose(wdf.corrwith(wdf.x), df.corrwith(df.x),
                    rtol=1e-12, atol=1e-12)


def test_mcmc_series_stats(mcmc_df, mcmc_wdf):
    sx = mcmc_df.loc[:, 'x']
    sy = mcmc_df.loc[:, 'y']
    wsx = mcmc_wdf.loc[:, 'x']
    wsy = mcmc_wdf.loc[:, 'y']
    assert_allclose(wsx.mean(), sx.mean(), rtol=1e-12)
    assert_allclose(wsx.var(ddof=0), sx.var(ddof=0), rtol=1e-12)
    assert_allclose(wsx.var(ddof=1), sx.var(ddof=1), rtol=1e-12)
    assert_allclose(wsx.std(ddof=0), sx.std(ddof=0), rtol=1e-12)
    assert_allclose(wsx.std(ddof=1), sx.std(ddof=1), rtol=1e-12)
    assert_allclose(wsx.sem(ddof=0), sx.sem(ddof=0), rtol=1e-12)
    assert_allclose(wsx.sem(ddof=1), sx.sem(ddof=1), rtol=1e-12)
    assert_allclose(wsx.skew(), sx.skew(), rtol=1e-12)
    assert_allclose(wsx.kurt(), sx.kurt(), rtol=1e-12)
    assert_allclose(wsx.cov(wsy, ddof=0), sx.cov(sy, ddof=0), rtol=1e-12)
    assert_allclose(wsx.cov(wsy, ddof=1), sx.cov(sy, ddof=1), rtol=1e-12)
    assert_allclose(wsx.cov(wsx, ddof=0), sx.var(ddof=0), rtol=1e-12)
    assert_allclose(wsx.cov(wsx, ddof=1), sx.var(ddof=1), rtol=1e-12)
    assert_allclose(wsx.corr(wsy), sx.corr(sy), rtol=1e-12, atol=1e-12)


@pytest.fixture
def df_small():
    np.random.seed(42)
    ncol = 3
    ndat = 10
    dat = np.random.normal(loc=5, scale=1, size=(ndat, ncol))
    return DataFrame(dat, columns=["x", "y", "z"])


@pytest.fixture
def wdf_small(df_small):
    np.random.seed(42)
    w = np.random.rand(df_small.shape[0])
    return WeightedDataFrame(df_small.to_numpy(), weights=w,
                             columns=df_small.columns)


def test_WeightedDataFrame_hist(mcmc_df, mcmc_wdf):
    df_axes = mcmc_df.hist().flatten()
    wdf_axes = mcmc_wdf.hist().flatten()

    df_heights = [p.get_height() for ax in df_axes for p in ax.patches]
    wdf_heights = [p.get_height() for ax in wdf_axes for p in ax.patches]
    assert df_heights == wdf_heights

    df_widths = [p.get_width() for ax in df_axes for p in ax.patches]
    wdf_widths = [p.get_width() for ax in wdf_axes for p in ax.patches]
    assert df_widths == wdf_widths

    df_xys = [p.get_xy() for ax in df_axes for p in ax.patches]
    wdf_xys = [p.get_xy() for ax in wdf_axes for p in ax.patches]
    assert df_xys == wdf_xys

    df_axes = mcmc_df.plot.hist(subplots=True)
    wdf_axes = mcmc_wdf.plot.hist(subplots=True)

    df_heights = [p.get_height() for ax in df_axes for p in ax.patches]
    wdf_heights = [p.get_height() for ax in wdf_axes for p in ax.patches]
    assert df_heights == wdf_heights

    df_widths = [p.get_width() for ax in df_axes for p in ax.patches]
    wdf_widths = [p.get_width() for ax in wdf_axes for p in ax.patches]
    assert df_widths == wdf_widths

    df_xys = [p.get_xy() for ax in df_axes for p in ax.patches]
    wdf_xys = [p.get_xy() for ax in wdf_axes for p in ax.patches]
    assert df_xys == wdf_xys


def test_WeightedSeries_hist(mcmc_df, mcmc_wdf):

    fig, axes = plt.subplots(2)
    mcmc_df.x.hist(ax=axes[0])
    mcmc_wdf.x.hist(ax=axes[1])

    df_heights = [p.get_height() for p in axes[0].patches]
    wdf_heights = [p.get_height() for p in axes[1].patches]
    assert df_heights == wdf_heights

    df_widths = [p.get_width() for p in axes[0].patches]
    wdf_widths = [p.get_width() for p in axes[1].patches]
    assert df_widths == wdf_widths

    df_xys = [p.get_xy() for p in axes[0].patches]
    wdf_xys = [p.get_xy() for p in axes[1].patches]
    assert df_xys == wdf_xys

    fig, axes = plt.subplots(2)
    mcmc_df.x.plot.hist(ax=axes[0])
    mcmc_wdf.x.plot.hist(ax=axes[1])

    df_heights = [p.get_height() for p in axes[0].patches]
    wdf_heights = [p.get_height() for p in axes[1].patches]
    assert df_heights == wdf_heights

    df_widths = [p.get_width() for p in axes[0].patches]
    wdf_widths = [p.get_width() for p in axes[1].patches]
    assert df_widths == wdf_widths

    df_xys = [p.get_xy() for p in axes[0].patches]
    wdf_xys = [p.get_xy() for p in axes[1].patches]
    assert df_xys == wdf_xys


def test_KdePlot(mcmc_df, mcmc_wdf):
    bw_method = 0.3
    fig, axes = plt.subplots(2)
    mcmc_df.x.plot.kde(bw_method=bw_method, ax=axes[0])
    mcmc_wdf.x.plot.kde(bw_method=bw_method, ax=axes[1])
    df_line, wdf_line = axes[0].lines[0], axes[1].lines[0]
    assert (df_line.get_xdata() == wdf_line.get_xdata()).all()
    assert_allclose(df_line.get_ydata(),  wdf_line.get_ydata(), atol=1e-4)


def test_scatter_matrix(mcmc_df, mcmc_wdf):
    axes = scatter_matrix(mcmc_df)
    data = axes[0, 1].collections[0].get_offsets().data
    axes = orig_scatter_matrix(mcmc_df)
    orig_data = axes[0, 1].collections[0].get_offsets().data

    assert_allclose(data, orig_data)

    axes = scatter_matrix(mcmc_wdf)
    data = axes[0, 1].collections[0].get_offsets().data
    n = len(data)
    n_ = neff(mcmc_wdf.get_weights(), beta='equal')
    assert_allclose(n, n_, atol=np.sqrt(n))

    axes = orig_scatter_matrix(mcmc_wdf)
    orig_data = axes[0, 1].collections[0].get_offsets().data
    n = len(orig_data)
    assert n == len(mcmc_wdf)

    axes = scatter_matrix(mcmc_wdf, ncompress=50)
    n = len(axes[0, 1].collections[0].get_offsets().data)
    assert_allclose(n, 50, atol=np.sqrt(50))


def test_bootstrap_plot(mcmc_df, mcmc_wdf):
    bootstrap_plot(mcmc_wdf.x)
    bootstrap_plot(mcmc_wdf.x, ncompress=500)


def test_BoxPlot(mcmc_df, mcmc_wdf):
    mcmc_df.plot.box()
    mcmc_wdf.plot.box()

    fig, ax = plt.subplots()
    mcmc_df.boxplot(ax=ax)

    fig, ax = plt.subplots()
    mcmc_wdf.boxplot(ax=ax)

    fig, ax = plt.subplots()
    mcmc_df.x.plot.box(ax=ax)

    fig, ax = plt.subplots()
    mcmc_wdf.x.plot.box(ax=ax)

    mcmc_df.plot.box(subplots=True)
    mcmc_wdf.plot.box(subplots=True)

    mcmc_df['split'] = ''
    mcmc_df.loc[:len(mcmc_df)//2, 'split'] = 'A'
    mcmc_df.loc[len(mcmc_df)//2:, 'split'] = 'B'

    mcmc_wdf['split'] = ''
    mcmc_wdf.iloc[:len(mcmc_wdf)//2, -1] = 'A'
    mcmc_wdf.iloc[len(mcmc_wdf)//2:, -1] = 'B'

    mcmc_df.groupby('split').plot.box()
    mcmc_wdf.groupby('split').plot.box()

    for return_type in ['dict', 'both']:
        fig, ax = plt.subplots()
        mcmc_wdf.plot.box(return_type=return_type, ax=ax)
        fig, ax = plt.subplots()
        mcmc_wdf.boxplot(return_type=return_type, ax=ax)

    fig, ax = plt.subplots()
    mcmc_wdf.boxplot(xlabel='xlabel', ax=ax)
    fig, ax = plt.subplots()
    mcmc_wdf.boxplot(ylabel='ylabel', ax=ax)

    fig, ax = plt.subplots()
    if version.parse(matplotlib.__version__) >= version.parse('3.10.0'):
        mcmc_wdf.boxplot(orientation='horizontal', ax=ax)
    else:
        mcmc_wdf.boxplot(vert=False, ax=ax)

    fig, ax = plt.subplots()
    mcmc_wdf.boxplot(fontsize=30, ax=ax)


def test_ScatterPlot(mcmc_df, mcmc_wdf):
    mcmc_df.plot.scatter('x', 'y')
    ax = mcmc_wdf.plot.scatter('x', 'y')

    n = len(ax.collections[0].get_offsets().data)
    n_ = neff(mcmc_wdf.get_weights(), beta='equal')
    assert_allclose(n, n_, atol=np.sqrt(n))

    ax = mcmc_wdf.plot.scatter('x', 'y', ncompress=50)
    n = len(ax.collections[0].get_offsets().data)
    assert_allclose(n, 50, atol=np.sqrt(50))


def test_HexBinPlot(mcmc_df, mcmc_wdf):
    df_axes = mcmc_df.plot.hexbin('x', 'y', mincnt=1)
    wdf_axes = mcmc_wdf.plot.hexbin('x', 'y')

    df_data = df_axes.collections[0].get_offsets()
    wdf_data = wdf_axes.collections[0].get_offsets()
    assert_allclose(df_data, wdf_data)

    df_colors = df_axes.collections[0].get_facecolors()
    wdf_colors = wdf_axes.collections[0].get_facecolors()
    assert_allclose(df_colors, wdf_colors)

    plt.close("all")


def test_AreaPlot(mcmc_df, mcmc_wdf):
    fig, (ax1, ax2) = plt.subplots(1, 2)
    axes_df = mcmc_df.plot.area(ax=ax1)
    axes_wdf = mcmc_wdf.plot.area(ax=ax2)

    assert_allclose(axes_df.get_xlim(), axes_wdf.get_xlim(), rtol=1e-3)
    assert_allclose(axes_df.get_ylim(), axes_wdf.get_ylim(), rtol=1e-3)


def test_BarPlot(mcmc_df, mcmc_wdf):
    axes_bar = mcmc_wdf[5:10].plot.bar()
    axes_barh = mcmc_wdf[5:10].plot.barh()
    assert_array_equal(axes_bar.get_xticks(), axes_barh.get_yticks())
    assert_array_equal(axes_bar.get_yticks(), axes_barh.get_xticks())


def test_PiePlot(mcmc_df, mcmc_wdf):
    fig, ax = plt.subplots()
    mcmc_wdf[5:10].x.plot.pie(ax=ax)
    mcmc_wdf[5:10].plot.pie(subplots=True)


def test_LinePlot(mcmc_df, mcmc_wdf):
    fig, ax = plt.subplots()
    df_axes = mcmc_df.x.plot.line(ax=ax)
    assert_array_equal(mcmc_df.index, df_axes.lines[0].get_xdata())
    fig, ax = plt.subplots()
    wdf_axes = mcmc_wdf.x.plot.line(ax=ax)
    assert_array_equal(mcmc_wdf.index.droplevel('weights'),
                       wdf_axes.lines[0].get_xdata())
    wdf_axes = mcmc_wdf.plot.line()
    assert len(wdf_axes.lines) == len(mcmc_wdf.columns)


def test_multiindex(mcmc_wdf):
    np.random.seed(0)
    i1 = np.arange(len(mcmc_wdf.index))
    i2 = np.random.randint(0, len(mcmc_wdf.index), len(mcmc_wdf.index))
    i3 = np.random.randint(0, len(mcmc_wdf.index), len(mcmc_wdf.index))
    index = MultiIndex.from_arrays([i1, i2, i3], names=['A', 'B', 'C'])
    weights = mcmc_wdf.get_weights()
    wdf = WeightedDataFrame(mcmc_wdf.values, weights=weights, index=index)

    assert wdf.index.names == ['A', 'B', 'C', 'weights']
    assert_allclose(np.array([*wdf.index]).T, [i1, i2, i3, weights])

    assert wdf.reset_index().index.names == [None, 'weights']

    assert not np.array_equal(wdf.reset_index().columns, wdf.columns)
    assert_array_equal(wdf.reset_index(drop=True).columns, wdf.columns)

    new = wdf.copy()
    assert not np.array_equal(new.index, wdf.reset_index())
    new.reset_index(inplace=True)
    assert_array_equal(new.index, wdf.reset_index().index)

    wdf_ = wdf.reset_index(level='A')
    assert wdf_.index.names == ['B', 'C', 'weights']
    wdf_ = wdf.reset_index(level=['A', 'C'])
    assert wdf_.index.names == ['B', 'weights']

    assert_array_equal(wdf.get_weights(), weights)
    wdf_ = wdf.reorder_levels(['B', 'C', 'weights', 'A'])
    assert_array_equal(wdf_.get_weights(), weights)
    weights_ = np.random.rand(len(weights))
    assert_array_equal(wdf_.set_weights(weights_).get_weights(), weights_)

    wdf_ = wdf.droplevel('weights').set_weights(weights, level=2)
    assert wdf_.index.names == ['A', 'B', 'weights', 'C']
    assert_array_equal(wdf_.get_weights(), weights)

    wdf_ = wdf.set_weights(weights, level=2)
    assert wdf_.index.names == ['A', 'B', 'weights', 'C']
    assert_array_equal(wdf_.get_weights(), weights)


def test_weight_passing(mcmc_wdf):
    weights = mcmc_wdf.get_weights().copy()
    new_wdf = WeightedDataFrame(mcmc_wdf.copy(), weights=None)
    assert (new_wdf.get_weights() == mcmc_wdf.get_weights()).all()

    np.random.shuffle(weights)
    assert (mcmc_wdf.get_weights() != weights).any()

    new_wdf = WeightedDataFrame(mcmc_wdf.copy(), weights=weights)
    assert_array_equal(new_wdf.get_weights(), weights)


def test_set_weights(mcmc_wdf):
    weights_1 = mcmc_wdf.get_weights()
    weights_2 = np.random.rand(len(mcmc_wdf.index))

    assert_array_equal(mcmc_wdf.set_weights(weights_2).get_weights(),
                       weights_2)
    assert_array_equal(mcmc_wdf.get_weights(), weights_1)
    mcmc_wdf.set_weights(weights_2, inplace=True)
    assert_array_equal(mcmc_wdf.get_weights(), weights_2)

    assert mcmc_wdf.isweighted()
    assert not mcmc_wdf.set_weights(None).isweighted()
    assert_array_equal(mcmc_wdf.set_weights(None).get_weights(), 1)

    mcmc_wdf.set_weights(None, inplace=True)
    assert not mcmc_wdf.isweighted()
    assert mcmc_wdf.set_weights(None) is not mcmc_wdf

    mcmc_id = id(mcmc_wdf)
    mcmc_wdf.set_weights(None, inplace=True)
    assert id(mcmc_wdf) == mcmc_id
    assert not mcmc_wdf.isweighted()


def test_drop_weights(mcmc_wdf):
    assert mcmc_wdf.isweighted()
    noweights = mcmc_wdf.drop_weights()
    assert not noweights.isweighted()
    assert_array_equal(noweights, mcmc_wdf)
    pandas.testing.assert_frame_equal(noweights.drop_weights(), noweights)
    assert noweights.drop_weights() is not noweights


def test_blank_axis_labels(df_small, wdf_small):
    for df in df_small, wdf_small:
        assert df.plot.area().get_xlabel() == ""
        assert df.plot.bar().get_xlabel() == ""
        assert df.plot.barh().get_ylabel() == ""


def test_get_index(wdf_small):
    wdf_small.index = wdf_small.index.rename(('foo', 'weights'))
    assert wdf_small.plot.area().get_xlabel() == "foo"
    assert wdf_small.plot.bar().get_xlabel() == "foo"
    assert wdf_small.plot.barh().get_ylabel() == "foo"
    for plot in [wdf_small.plot.area, wdf_small.plot.bar, wdf_small.plot.barh]:
        assert plot(xlabel="my xlabel").get_xlabel() == "my xlabel"

    idx = wdf_small.index.to_frame().to_numpy().T
    wdf_small.index = MultiIndex.from_arrays([idx[0], idx[0], idx[1]],
                                             names=('foo', 'bar', 'weights'))
    assert wdf_small.plot.bar().get_xlabel() == "foo,bar"
    assert wdf_small.plot.barh().get_ylabel() == "foo,bar"
    wdf_small.index = MultiIndex.from_arrays([idx[0], idx[0], idx[1]],
                                             names=(None, None, 'weights'))
    assert wdf_small.plot.bar().get_xlabel() == ""
    assert wdf_small.plot.barh().get_ylabel() == ""


def test_style(mcmc_wdf):
    style_dict = dict(x='c', y='y', z='m', w='k')
    ax = mcmc_wdf.plot.kde(style=style_dict)
    assert all([line.get_color() == to_rgba(c)
                for line, c in zip(ax.get_lines(), ['c', 'y', 'm', 'k'])])

    with pytest.raises(TypeError):
        ax = mcmc_wdf.plot.hist_2d('x', 'y', style='c')
    with pytest.raises(TypeError):
        ax = mcmc_wdf.plot.kde_2d('x', 'y', style='c')
    with pytest.raises(TypeError):
        ax = mcmc_wdf.plot.scatter_2d('x', 'y', style='c')


def test_read_csv(tmp_path, mcmc_wdf):
    filename = tmp_path / 'mcmc_wdf.csv'

    mcmc_wdf.to_csv(filename)
    mcmc_wdf_ = read_csv(filename)
    pandas.testing.assert_frame_equal(mcmc_wdf, mcmc_wdf_)

    mcmc_wdf.set_weights(np.random.rand(mcmc_wdf.shape[1]),
                         axis=1, inplace=True)
    mcmc_wdf.to_csv(filename)
    mcmc_wdf_ = read_csv(filename)
    pandas.testing.assert_frame_equal(mcmc_wdf, mcmc_wdf_)

    mcmc_wdf = mcmc_wdf.drop_weights(axis=0)
    mcmc_wdf.to_csv(filename)
    mcmc_wdf_ = read_csv(filename)
    pandas.testing.assert_frame_equal(mcmc_wdf, mcmc_wdf_)

    mcmc_wdf = mcmc_wdf.drop_weights(axis=1)
    mcmc_wdf.to_csv(filename)
    mcmc_wdf_ = read_csv(filename)
    pandas.testing.assert_frame_equal(mcmc_wdf, mcmc_wdf_)



"""Read and write CSV files for anesthetic."""
from anesthetic.weighted_labelled_pandas import read_csv as wl_read_csv
from anesthetic.samples import MCMCSamples, NestedSamples
from pathlib import Path


def read_csv(filename, *args, **kwargs):
    """Read a CSV file into a :class:`anesthetic.samples.Samples` object."""
    try:
        filename = Path(filename)
        kwargs['label'] = kwargs.get('label', filename.stem)
        filename = filename.with_suffix('.csv')
    except TypeError:
        pass
    wldf = wl_read_csv(filename)
    if 'nlive' in wldf.columns:
        return NestedSamples(wldf, *args, **kwargs)
    else:
        return MCMCSamples(wldf, *args, **kwargs)
